"""Kaskada query builder and local execution engine."""
from __future__ import annotations

from . import destinations, plot, results, sources, windows
from ._execution import Execution, ResultIterator
from ._session import init_session
from ._timestream import Arg, LiteralValue, Timestream, record
from .udf import udf


__all__ = [
    "Arg",
    "destinations",
    "Execution",
    "init_session",
    "LiteralValue",
    "plot",
    "record",
    "ResultIterator",
    "results",
    "sources",
    "Timestream",
    "udf",
    "windows",
]



"""Provide sources based on PyArrow, including Pandas and CSV."""
from __future__ import annotations

from io import BytesIO
from typing import Optional

import pandas as pd
import pyarrow as pa
import pyarrow.csv
import pyarrow.json
import pyarrow.parquet

from .. import _ffi
from .._session import _get_session
from .source import Source, TimeUnit


class Pandas(Source):
    """Source reading data from Pandas dataframe."""

    def __init__(
        self,
        *,
        time_column: str,
        key_column: str,
        schema: pa.Schema,
        subsort_column: Optional[str] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> None:
        """Create a source reading Pandas DataFrames.

        Args:
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            schema: The schema to use.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        super().__init__(
            schema=schema,
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )

    @staticmethod
    async def create(
        dataframe: Optional[pd.DataFrame] = None,
        *,
        time_column: str,
        key_column: str,
        subsort_column: Optional[str] = None,
        schema: Optional[pa.Schema] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> Pandas:
        """Create a source reading Pandas DataFrames.

        Args:
            dataframe: The DataFrame to start from.
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            schema: The schema to use. If not provided, it will be inferred from the input.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        if schema is None:
            if dataframe is None:
                raise ValueError("Must provide schema or dataframe")
            schema = pa.Schema.from_pandas(dataframe)

        source = Pandas(
            schema=schema,
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )

        if dataframe is not None:
            await source.add_data(dataframe)
        return source

    async def add_data(self, data: pd.DataFrame) -> None:
        """Add data to the source."""
        table = pa.Table.from_pandas(data, self._schema, preserve_index=False)
        for batch in table.to_batches():
            await self._ffi_table.add_pyarrow(batch)


class PyDict(Source):
    """Source reading data from lists of dicts."""

    def __init__(
        self,
        *,
        time_column: str,
        key_column: str,
        schema: pa.Schema,
        queryable: bool = True,
        subsort_column: Optional[str] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> None:
        """Create a source reading from rows represented as dicts.

        Args:
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            schema: The schema to use.
            queryable: Whether added rows will be available for running queries.
              If True, rows (both provided to the constructor and added later) will be available
              for interactive queries. If False, rows will be discarded after being sent to any
              running materializations. Consider setting this to False when the source will only
              be used for materialization to avoid unnecessary memory usage.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        super().__init__(
            queryable=queryable,
            schema=schema,
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )

        self._convert_options = pyarrow.csv.ConvertOptions(column_types=schema)

    @staticmethod
    async def create(
        rows: Optional[dict | list[dict]] = None,
        *,
        time_column: str,
        key_column: str,
        queryable: bool = True,
        subsort_column: Optional[str] = None,
        schema: Optional[pa.Schema] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> PyDict:
        """Create a source reading from rows represented as dicts.

        Args:
            rows: The input row(s) as a dictionary or list of dictionaries.
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            queryable: Whether added rows will be available for running queries.
              If True, rows (both provided to the constructor and added later) will be available
              for interactive queries. If False, rows will be discarded after being sent to any
              running materializations. Consider setting this to False when the source will only
              be used for materialization to avoid unnecessary memory usage.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            schema: The schema to use. If not provided, it will be inferred from the input.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        if schema is None:
            if rows is None:
                raise ValueError("Must provide schema or rows")
            schema = pa.Table.from_pylist(rows).schema
        source = PyDict(
            time_column=time_column,
            key_column=key_column,
            queryable=queryable,
            subsort_column=subsort_column,
            schema=schema,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )
        if rows:
            await source.add_rows(rows)
        return source

    async def add_rows(self, rows: dict | list[dict]) -> None:
        """Add data to the source."""
        if isinstance(rows, dict):
            rows = [rows]
        table = pa.Table.from_pylist(rows, schema=self._schema)
        for batch in table.to_batches():
            await self._ffi_table.add_pyarrow(batch)


# TODO: We should be able to go straight from CSV to PyArrow, but
# currently that has some problems with timestamp handling.
class CsvString(Source):
    """Source reading data from CSV strings using Pandas."""

    def __init__(
        self,
        *,
        time_column: str,
        key_column: str,
        schema: pa.Schema,
        subsort_column: Optional[str] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> None:
        """Create a CSV String Source.

        Args:
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            schema: The schema to use.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        super().__init__(
            schema=schema,
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )

        self._convert_options = pyarrow.csv.ConvertOptions(
            column_types=schema,
            strings_can_be_null=True,
        )

    _parse_options = pyarrow.csv.ParseOptions(
        escape_char="\\",
    )

    @staticmethod
    async def create(
        csv_string: Optional[str | BytesIO] = None,
        *,
        time_column: str,
        key_column: str,
        subsort_column: Optional[str] = None,
        schema: Optional[pa.Schema] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> CsvString:
        """Create a CSV String Source with data.

        Args:
            csv_string: The CSV string to start from.
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            schema: The schema to use. If not provided, it will be inferred from the input.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        if isinstance(csv_string, str):
            csv_string = BytesIO(csv_string.encode("utf-8"))
        if schema is None:
            if csv_string is None:
                raise ValueError("Must provide schema or csv_string")
            schema = pa.csv.read_csv(
                csv_string, parse_options=CsvString._parse_options
            ).schema
            csv_string.seek(0)

        source = CsvString(
            schema=schema,
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )

        if csv_string:
            await source.add_string(csv_string)
        return source

    async def add_string(self, csv_string: str | BytesIO) -> None:
        """Add data to the source."""
        if isinstance(csv_string, str):
            csv_string = BytesIO(csv_string.encode("utf-8"))
        content = pa.csv.read_csv(
            csv_string,
            convert_options=self._convert_options,
            parse_options=CsvString._parse_options,
        )
        for batch in content.to_batches():
            await self._ffi_table.add_pyarrow(batch)


class JsonlFile(Source):
    """Source reading data from line-delimited JSON files using PyArrow."""

    def __init__(
        self,
        *,
        time_column: str,
        key_column: str,
        schema: pa.Schema,
        subsort_column: Optional[str] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> None:
        """Create a line-delimited JSON File Source.

        Args:
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            schema: The schema to use.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        super().__init__(
            schema=schema,
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )
        self._parse_options = pyarrow.json.ParseOptions(explicit_schema=schema)

    @staticmethod
    async def create(
        path: Optional[str] = None,
        *,
        time_column: str,
        key_column: str,
        subsort_column: Optional[str] = None,
        schema: Optional[pa.Schema] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> JsonlFile:
        """Create a source reading a line-delimited JSON file.

        Args:
            path: The path to the line-delimited JSON file to add. This can be relative to
              the current working directory or an absolute path (prefixed by '/').
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            schema: The schema to use. If not provided, it will be inferred from the input.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        path = Source._get_absolute_path(path)

        if schema is None:
            if path is None:
                raise ValueError("Must provide schema or path to jsonl file")
            schema = pa.json.read_json(path).schema

        source = JsonlFile(
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            schema=schema,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )

        if path:
            await source.add_file(path)
        return source

    async def add_file(self, path: str) -> None:
        """Add data to the source."""
        batches = pa.json.read_json(
            Source._get_absolute_path(path), parse_options=self._parse_options
        )
        for batch in batches.to_batches():
            await self._ffi_table.add_pyarrow(batch)


class JsonlString(Source):
    """Source reading data from line-delimited JSON strings using PyArrow."""

    def __init__(
        self,
        *,
        time_column: str,
        key_column: str,
        schema: pa.Schema,
        subsort_column: Optional[str] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> None:
        """Create a JSON String Source.

        Args:
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            schema: The schema to use.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        super().__init__(
            schema=schema,
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )
        self._parse_options = pyarrow.json.ParseOptions(explicit_schema=schema)

    @staticmethod
    async def create(
        json_string: Optional[str | BytesIO] = None,
        *,
        time_column: str,
        key_column: str,
        subsort_column: Optional[str] = None,
        schema: Optional[pa.Schema] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> JsonlString:
        """Create a source reading from JSON strings.

        Args:
            json_string: The line-delimited JSON string to start from.
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            schema: The schema to use. If not provided, it will be inferred from the input.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        if isinstance(json_string, str):
            json_string = BytesIO(json_string.encode("utf-8"))
        if schema is None:
            if json_string is None:
                raise ValueError("Must provide schema or JSON")
            schema = pa.json.read_json(json_string).schema
            json_string.seek(0)

        source = JsonlString(
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            schema=schema,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )

        if json_string:
            await source.add_string(json_string)
        return source

    async def add_string(self, json_string: str | BytesIO) -> None:
        """Add data to the source."""
        if isinstance(json_string, str):
            json_string = BytesIO(json_string.encode("utf-8"))
        batches = pa.json.read_json(json_string, parse_options=self._parse_options)
        for batch in batches.to_batches():
            await self._ffi_table.add_pyarrow(batch)


class Parquet(Source):
    """Source reading data from Parquet files."""

    def __init__(
        self,
        *,
        time_column: str,
        key_column: str,
        schema: Optional[pa.Schema],
        subsort_column: Optional[str] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> None:
        """Create a Parquet source.

        Args:
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            schema: The schema to use.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        super().__init__(
            schema=schema,
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            grouping_name=grouping_name,
            time_unit=time_unit,
            source="parquet",
        )

    @staticmethod
    async def create(
        file: Optional[str] = None,
        *,
        time_column: str,
        key_column: str,
        schema: Optional[pa.Schema] = None,
        subsort_column: Optional[str] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
    ) -> Parquet:
        """Create a Parquet source.

        Args:
            file: The url or path of the Parquet file to add. Paths should be relative to the
              current working directory or absolute. URLs may describe local file paths or
              object-store locations.
            time_column: The name of the column containing the time.
            key_column: The name of the column containing the key.
            schema: The schema to use. If not provided, it will be inferred from the input.
            subsort_column: The name of the column containing the subsort.
              If not provided, the subsort will be assigned by the system.
            grouping_name: The name of the group associated with each key.
              This is used to ensure implicit joins are only performed between data grouped
              by the same entity.
            time_unit: The unit of the time column. One of `ns`, `us`, `ms`, or `s`.
              If not specified (and not specified in the data), nanosecond will be assumed.
        """
        if schema is None:
            if file is None:
                raise ValueError("Must provide schema or url to parquet file")
            schema = await _ffi.parquet_schema(_get_session(), file)
        source = Parquet(
            time_column=time_column,
            key_column=key_column,
            subsort_column=subsort_column,
            schema=schema,
            grouping_name=grouping_name,
            time_unit=time_unit,
        )

        if file:
            await source.add_file(file)
        return source

    async def add_file(self, file: str) -> None:
        """Add data to the source."""
        await self._ffi_table.add_parquet(file)



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def count_if_source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n,is_valid",
            "1996-12-19T16:39:57,A,5,10,true",
            "1996-12-19T16:39:58,B,24,3,true",
            "1996-12-19T16:39:59,A,17,6,false",
            "1996-12-19T16:40:00,A,,9,false",
            "1996-12-19T16:40:01,A,12,,true",
            "1996-12-19T16:40:02,A,,,",
            "1996-12-19T16:40:03,B,26,12,true",
            "1996-12-19T16:40:04,B,30,1,true",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_count_if_unwindowed(count_if_source, golden) -> None:
    is_valid = count_if_source.col("is_valid")
    m = count_if_source.col("m")
    golden.jsonl(
        kd.record(
            {
                "is_valid": is_valid,
                "count_if": is_valid.count_if(),
                "m": m,
            }
        )
    )


async def test_count_if_windowed(count_if_source, golden) -> None:
    is_valid = count_if_source.col("is_valid")
    m = count_if_source.col("m")
    golden.jsonl(
        kd.record(
            {
                "is_valid": is_valid,
                "since_count_if": is_valid.count_if(window=kd.windows.Since(m > 25)),
                "m": m,
            }
        )
    )


async def test_count_if_since_true(count_if_source, golden) -> None:
    is_valid = count_if_source.col("is_valid")
    m = count_if_source.col("m")
    golden.jsonl(
        kd.record(
            {
                "is_valid": is_valid,
                "since_count_if": is_valid.count_if(window=kd.windows.Since(True)),
                "m": m,
            }
        )
    )



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1996-12-19T16:40:00,A,,9",
            "1996-12-19T16:40:01,A,12,",
            "1996-12-19T16:40:02,A,,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


@pytest.fixture(scope="module")
async def source_spread_across_days() -> kd.sources.PyDict:
    return await kd.sources.PyDict.create(
        [
            {"time": "2021-01-01T00:00:00", "key": "A", "m": 1, "n": 2},
            {"time": "2021-01-01T01:10:01", "key": "A", "m": 3, "n": 4},
            {"time": "2021-01-01T02:20:02", "key": "A", "m": 5, "n": 6},
            {"time": "2021-01-01T03:20:03", "key": "A", "m": 7, "n": 8},
            {"time": "2021-01-01T04:35:04", "key": "A", "m": 9, "n": 10},
            {"time": "2021-01-01T05:40:05", "key": "A", "m": 11, "n": 12},
            {"time": "2021-01-01T05:45:06", "key": "A", "m": 13, "n": 14},
        ],
        time_column="time",
        key_column="key",
    )


async def test_sum_unwindowed(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(kd.record({"m": m, "sum_m": m.sum(), "n": n, "sum_n": n.sum()}))


async def test_sum_windowed(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "since_sum_m": m.sum(window=kd.windows.Since(m > 20)),
                "n": n,
                "sliding_sum_n": n.sum(window=kd.windows.Sliding(2, m > 10)),
            }
        )
    )


async def test_sum_since_true(source, golden) -> None:
    # `since(True)` should be the same as unwindowed, so equals the original vaule.
    m_sum_since_true = kd.record(
        {
            "m": source.col("m"),
            "since_m_sum": source.col("m").sum(window=kd.windows.Since(True)),
        }
    )
    golden.jsonl(m_sum_since_true)


async def test_sum_since_hourly(source_spread_across_days, golden) -> None:
    golden.jsonl(
        source_spread_across_days.col("m").sum(window=kd.windows.Since.hourly())
    )


async def test_sum_tumbling_hourly(source_spread_across_days, golden) -> None:
    m = source_spread_across_days.col("m")
    golden.jsonl(
        kd.record({"tumbling_hourly": m.sum(window=kd.windows.Tumbling.hourly())})
    )


async def test_sum_tumbling_predicate(source_spread_across_days, golden) -> None:
    m = source_spread_across_days.col("m")
    golden.jsonl(kd.record({"f1": m.sum(window=kd.windows.Tumbling(m > 8))}))



import kaskada as kd
import pyarrow as pa
import pytest


def test_table_valid() -> None:
    schema = pa.schema(
        [
            pa.field("time", pa.int32(), nullable=False),
            pa.field("key", pa.int64(), nullable=False),
        ]
    )

    kd.sources.Source(schema, time_column="time", key_column="key")


def test_table_invalid_names() -> None:
    schema = pa.schema(
        [
            pa.field("time", pa.int32(), nullable=False),
            pa.field("key", pa.int64(), nullable=False),
        ]
    )

    with pytest.raises(KeyError):
        # Currently, this doesn't propagate the suggestions from
        # existing column names from Sparrow.
        # TODO: Do that.
        kd.sources.Source(schema, time_column="non_existant_time", key_column="key")

    with pytest.raises(KeyError):
        # Currently, this doesn't propagate the suggestions from
        # existing column names from Sparrow.
        # TODO: Do that.
        kd.sources.Source(schema, time_column="time", key_column="non_existant_key")

    with pytest.raises(KeyError):
        # Currently, this doesn't propagate the suggestions from
        # existing column names from Sparrow.
        # TODO: Do that.
        kd.sources.Source(
            schema,
            time_column="time",
            key_column="key",
            subsort_column="non_existant_subsort",
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import AsyncIterator, Callable, Iterator, Literal, Optional, TypeVar, Union

import kaskada as kd
import pyarrow as pa

from . import _ffi


T = TypeVar("T")


@dataclass
class _ExecutionOptions:
    """Execution options passed to the FFI layer."""

    #: The maximum number of rows to return.
    #: If not specified, all rows are returned.
    row_limit: Optional[int] = None

    #: The maximum batch size to use when returning results.
    #: If not specified, the default batch size will be used.
    max_batch_size: Optional[int] = None

    #: If true, the query will be a continuous materialization.
    materialize: bool = False

    #: The type of results to return.
    results: Literal["history", "snapshot"] = "history"

    #: The earliest time of changes to include in the results.
    #: For history, this limits the points output.
    #: For snapshots, this limits the entities that are output.
    changed_since: Optional[int] = None

    #: The last time to process.
    #: If not set, defaults to the current time.
    #: For history, this limits the points output.
    #: For snapshots, this determines when the snapshot is produced.
    final_at: Optional[int] = None

    @staticmethod
    def create(
        results: Optional[Union[kd.results.History, kd.results.Snapshot]],
        row_limit: Optional[int],
        max_batch_size: Optional[int],
        mode: Literal["once", "live"] = "once",
    ) -> _ExecutionOptions:
        """Create execution options."""
        options = _ExecutionOptions(
            row_limit=row_limit,
            max_batch_size=max_batch_size,
            materialize=mode == "live",
        )

        if results is None:
            results = kd.results.History()

        if isinstance(results, kd.results.History):
            options.results = "history"
            if results.since is not None:
                options.changed_since = int(results.since.timestamp())
            if results.until is not None:
                options.final_at = int(results.until.timestamp())
        elif isinstance(results, kd.results.Snapshot):
            options.results = "snapshot"
            if results.changed_since is not None:
                options.changed_since = int(results.changed_since.timestamp())
            if results.at is not None:
                options.final_at = int(results.at.timestamp())
        else:
            raise AssertionError(f"Unhandled results type {results!r}")
        return options


class Execution:
    """Represents an execution of a TimeStream."""

    def __init__(self, ffi_execution: _ffi.Execution) -> None:
        """Create the execution for a given FFI call."""
        self._ffi_execution = ffi_execution

    def stop(self):
        """Stop or cancel the execution.

        After this call, the execution will no longer produce results.
        """
        self._ffi_execution.stop()


class ResultIterator(Execution, Iterator[T], AsyncIterator[T]):
    """An iterator over results from the TimeStream."""

    _items: Iterator[T]

    def __init__(
        self, ffi_execution: _ffi.Execution, f: Callable[[pa.Table], Iterator[T]]
    ) -> None:
        """Create the execution for a given FFI call."""
        super().__init__(ffi_execution)
        self._f = f
        self._items = iter(())

    def __iter__(self):
        """Return a synchronous iterator over results."""
        return self

    def __aiter__(self):
        """Return an asynchronous iterator over results."""
        return self

    def __next__(self) -> T:
        """Return the next item synchronously."""
        try:
            return next(self._items)
        except StopIteration:
            pass

        while True:
            next_batch = self._ffi_execution.next_pyarrow()
            if next_batch is None:
                raise StopIteration

            # Annoyingly, PyArrow doesn't suport `drop_columns` on batches.
            # So we need to convert to a Table (even if we're producing batches).
            table = pa.Table.from_batches([next_batch])
            table = table.drop_columns(["_subsort", "_key_hash"])
            self._items = self._f(table)
            try:
                return next(self._items)
            except StopIteration:
                continue

    async def __anext__(self) -> T:
        """Return the next item asynchronously."""
        try:
            return next(self._items)
        except StopIteration:
            pass

        while True:
            next_batch = await self._ffi_execution.next_pyarrow_async()
            if next_batch is None:
                raise StopAsyncIteration

            # Annoyingly, PyArrow doesn't suport `drop_columns` on batches.
            # So we need to convert to a Table (even if we're producing batches).
            table = pa.Table.from_batches([next_batch])
            table = table.drop_columns(["_subsort", "_key_hash"])
            self._items = self._f(table)

            try:
                return next(self._items)
            except StopIteration:
                continue



"""Functionality for calling Python UDFs from Kaskada."""
from __future__ import annotations

import functools
from typing import Callable

import pandas as pd
import pyarrow as pa

from . import _ffi
from ._timestream import Arg, Timestream


# TODO: Allow functions to return `pd.DataFrame` for struct arrays.
FuncType = Callable[..., pd.Series]


class Udf:
    """User defined function."""

    def __init__(self, signature: str, func: Callable[..., pa.Array]) -> None:
        """Create a UDF with the given signature."""
        self._ffi = _ffi.Udf(signature, func)

    def __call__(self, *args: Arg) -> Timestream:
        """Apply the UDF to the given arguments."""
        return Timestream._call(self._ffi, *args)


def udf(signature: str):
    """Decorate a function for use as a Kaskada UDF."""

    def decorator(user_func: FuncType):
        # 1. Convert the `FuncType` to the type expected by Udf.
        # This needs to take the PyArrow result type and PyArrow inputs,
        # convert them to Pandas, call the function, and convert the result
        # back to PyArrow.
        def func(result_type, *args):
            return _converted_func(user_func, result_type, *args)

        # 2. Create the UDF object.
        udf = Udf(signature, func)

        # 3. Update the udf with the user function's documentation
        functools.update_wrapper(udf, user_func)

        return udf

    return decorator


def _converted_func(
    user_func: FuncType, result_type: pa.DataType, *args: pa.Array
) -> pa.Array:
    """Run the function producing the given result type."""
    # TODO: I believe this will return a series for simple arrays, and a
    # dataframe for struct arrays. We should explore how this handles
    # different types.
    pd_args = [arg.to_pandas() for arg in args]
    pd_result = user_func(*pd_args)

    if isinstance(pd_result, pd.Series):
        return pa.Array.from_pandas(pd_result, type=result_type)
    else:
        raise TypeError(f"Unsupported result type: {type(pd_result)}")



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1996-12-19T16:40:00,A,,9",
            "1996-12-19T16:40:01,A,12,",
            "1996-12-19T16:40:02,A,,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_if_(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    condition_m = m > 15
    condition_n = n > 5
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "condition_m": condition_m,
                "if_m": m.if_(condition_m),
                "n": n,
                "condition_n": condition_n,
                "if_n": n.if_(condition_n),
            }
        )
    )



import kaskada as kd
import kaskada._ffi as _ffi
import pandas as pd
import pytest
from kaskada.udf import Udf


@kd.udf("add<N: number>(x: N, y: N) -> N")
def add(x: pd.Series, y: pd.Series) -> pd.Series:
    """Use Pandas to add two numbers."""
    return x + y


@kd.udf("add_x2<N: number>(x: N, y: N) -> N")
def add_x2(x: pd.Series, y: pd.Series) -> pd.Series:
    """Use Pandas to add then multiply by 2"""
    return (x + y) * 2


def test_udf_instance() -> None:
    assert isinstance(add, Udf)
    assert isinstance(add._ffi, _ffi.Udf)


def test_docstring() -> None:
    assert add.__doc__ == "Use Pandas to add two numbers."


@pytest.fixture
async def source_int64() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1996-12-19T16:40:00,A,,9",
            "1996-12-19T16:40:01,A,12,",
            "1996-12-19T16:40:02,A,,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_add_udf_direct(golden, source_int64) -> None:
    m = source_int64.col("m")
    n = source_int64.col("n")
    golden.jsonl(kd.record({"m": m, "n": n, "add": add(m, n), "add_x2": add_x2(m, n)}))


async def test_add_udf_pipe(golden, source_int64) -> None:
    m = source_int64.col("m")
    n = source_int64.col("n")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "n": n,
                "add": m.pipe(add, n),
                "add_x2": m.pipe(add_x2, n),
            }
        )
    )


# Tests that redefining a UDF with the same signature overwrites the
# existing implementation.
#
# A bug existed where the UDF was hashing the signature rather than
# the uuid, so the DFG was equating two different UDFs with the
# same signature and using the first definition.
async def test_add_udf_redefined_inline(golden, source_int64) -> None:
    m = source_int64.col("m")
    n = source_int64.col("n")

    @kd.udf("my_sub<N: number>(x: N, y: N) -> N")
    def my_sub(x: pd.Series, y: pd.Series) -> pd.Series:
        return x - y

    applied_1 = my_sub(m, n)

    @kd.udf("my_sub<N: number>(x: N, y: N) -> N")  # type: ignore
    def my_sub(x: pd.Series, y: pd.Series) -> pd.Series:
        return (x - y) - 5

    applied_2 = my_sub(m, n)

    golden.jsonl(
        kd.record(
            {
                "m": m,
                "n": n,
                "rust_sub": m.sub(n),
                "udf_sub_1": applied_1,
                "udf_sub_2": applied_2,
            }
        )
    )



"""Defines methods for initializing the Kaskada session."""

from typing import Optional

from kaskada import _ffi


_SESSION: Optional[_ffi.Session] = None


def init_session() -> None:
    """Initialize the Kaskada session for this Python process.

    This must only be called once per session. It must be called before
    any other Kaskada functions are called.

    Raises:
        RuntimeError: If the session has already been initialized.
    """
    global _SESSION
    if _SESSION is not None:
        raise RuntimeError("Session has already been initialized")
    _SESSION = _ffi.Session()


def _get_session() -> _ffi.Session:
    """Assert that the session has been initialized.

    Returns: The FFI session handle.

    Raises:
        AssertionError: If the session has not been initialized.
    """
    global _SESSION
    assert _SESSION is not None, "Session has not been initialized"
    return _SESSION



"""Destinations for writing query results directly to external stores."""


class Destination(object):
    """Base class for destinations."""


class ParquetDestination(Destination):
    """Destination for Parquet files."""

    def __init__(self, path: str):
        """Construct a new ParquetDestination.

        Args:
            path: The path to the directory to write Parquet files.
        """
        raise NotImplementedError



"""Windows to use for Timestream aggregations."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from typing import Callable

from ._timestream import Timestream


@dataclass(frozen=True)
class Window(object):
    """Base class for window functions."""


@dataclass(frozen=True)
class Since(Window):
    """Window producing cumulative values since a predicate.

    "Since" windows are a series of non-overlapping windows producing cumulative values
    since each time the predicate evaluates to true. Each window is exclusive of the time
    it starts at and inclusive of the end time.

    Args:
        predicate: the condition used to determine when the window resets.

    Returns:
        Window for aggregating cumulative values since the predicate.
    """

    #: The boolean Timestream to use as predicate for the window.
    #: Each time the predicate evaluates to true the window will be cleared.
    #:
    #: The predicate may be a callable which returns the boolean Timestream, in
    #: which case it is applied to the Timestream being aggregated.
    predicate: Timestream | Callable[..., Timestream] | bool

    @staticmethod
    def minutely() -> Since:
        """Return a window since the start of each calendar minute."""
        return Since(predicate=lambda domain: Timestream._call("minutely", domain))

    @staticmethod
    def hourly() -> Since:
        """Return a window since the start of each calendar hour."""
        return Since(predicate=lambda domain: Timestream._call("hourly", domain))

    @staticmethod
    def daily() -> Since:
        """Return a window since the start of each calendar day."""
        return Since(predicate=lambda domain: Timestream._call("daily", domain))

    @staticmethod
    def monthly() -> Since:
        """Return a window since the start of each calendar month."""
        return Since(predicate=lambda domain: Timestream._call("monthly", domain))

    @staticmethod
    def yearly() -> Since:
        """Return a window since the start of each calendar year."""
        return Since(predicate=lambda domain: Timestream._call("yearly", domain))


@dataclass(frozen=True)
class Tumbling(Window):
    """Window producing a single value at each predicate.

    "Tumbling" windows are a series of non-overlapping windows that produce values
    each time the predicate evaluates to true. Each window is exclusive of the time
    it starts at and inclusive of the end time.

    Args:
        predicate: the condition used to determine when the window resets.

    Returns:
        Window for aggregating values since the predicate.

    Note:
        Like other systems, Kaskada treats tumbling windows as non-overlapping.
        When one window ends the next starts.

        Unlike other systems, tumbling windows do not need to be of a fixed size.
        Instead, they may be determined by a fixed duration, a calendar duration (such as a "month"),
        or even a predicate.
    """

    #: The boolean Timestream to use as predicate for the window.
    #: Each time the predicate evaluates to true the window will be cleared.
    #:
    #: The predicate may be a callable which returns the boolean Timestream, in
    #: which case it is applied to the Timestream being aggregated.
    predicate: Timestream | Callable[..., Timestream] | bool

    @staticmethod
    def minutely() -> Tumbling:
        """Return a tumbling window that resets at the start of each calendar minute."""
        return Tumbling(predicate=lambda domain: Timestream._call("minutely", domain))

    @staticmethod
    def hourly() -> Tumbling:
        """Return a tumbling window that resets at the start of each calendar hour."""
        return Tumbling(predicate=lambda domain: Timestream._call("hourly", domain))

    @staticmethod
    def daily() -> Tumbling:
        """Return a tumbling window that resets at the start of each calendar day."""
        return Tumbling(predicate=lambda domain: Timestream._call("daily", domain))

    @staticmethod
    def monthly() -> Tumbling:
        """Return a tumbling window that resets at the start of each calendar month."""
        return Tumbling(predicate=lambda domain: Timestream._call("monthly", domain))

    @staticmethod
    def yearly() -> Tumbling:
        """Return a tumbling window that resets at the start of each calendar year."""
        return Tumbling(predicate=lambda domain: Timestream._call("yearly", domain))


@dataclass(frozen=True)
class Sliding(Window):
    """Overlapping windows producing the latest value at each predicate.

    "Sliding" windows are a series of overlapping windows that produce values each
    time the predicate evaluates to true. Each window is exclusive of the time
    it starts at and inclusive of the end time.

    Args:
        duration: the number of active windows at any given time.
        predicate: the condition used to determine when the oldest window ends and a new
          window starts.

    Returns:
        Overlapping windows for aggregating values.
    """

    #: The number of sliding intervals to use in the window.
    duration: int
    #: The boolean Timestream to use as a predicate for the window.
    #: Each time the predicate evaluates to true the window starts a new interval.
    #:
    #: The predicate may be a callable which returns the boolean Timestream, in
    #: which case it is applied to the Timestream being aggregated.
    predicate: Timestream | Callable[..., Timestream] | bool

    def __post_init__(self):
        """Validate the window parameters."""
        if self.duration <= 0:
            raise ValueError("duration must be positive")

    @staticmethod
    def minutely(duration: int) -> Sliding:
        """Return a sliding window containing `duration` calendar minutes.

        Args:
            duration: The number of minutes to use in the window.
        """
        return Sliding(
            duration=duration,
            predicate=lambda domain: Timestream._call("minutely", domain),
        )

    @staticmethod
    def hourly(duration: int) -> Sliding:
        """Return a sliding window containing `duration` calendar hours.

        Args:
            duration: The number of hours to use in the window.
        """
        return Sliding(
            duration=duration,
            predicate=lambda domain: Timestream._call("hourly", domain),
        )

    @staticmethod
    def daily(duration: int) -> Sliding:
        """Return a sliding window containing `duration` calendar days.

        Args:
            duration: The number of days to use in the window.
        """
        return Sliding(
            duration=duration,
            predicate=lambda domain: Timestream._call("daily", domain),
        )

    @staticmethod
    def monthly(duration: int) -> Sliding:
        """Return a sliding window containing `duration` calendar months.

        Args:
            duration: The number of months to use in the window.
        """
        return Sliding(
            duration=duration,
            predicate=lambda domain: Timestream._call("monthly", domain),
        )

    @staticmethod
    def yearly(duration: int) -> Sliding:
        """Return a sliding window containing `duration` calendar years.

        Args:
            duration: The number of years to use in the window.
        """
        return Sliding(
            duration=duration,
            predicate=lambda domain: Timestream._call("yearly", domain),
        )


@dataclass(frozen=True)
class Trailing(Window):
    """Window the last `duration` time period.

    Args:
        duration: The duration of the window.
    """

    #: The duration of the window.
    duration: timedelta

    def __post_init__(self):
        """Validate the window parameters."""
        if self.duration <= timedelta(0):
            raise ValueError("duration must be positive")



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,new_key",
            "1996-12-19T16:39:57,A,5,C",
            "1996-12-19T16:39:58,B,24,D",
            "1996-12-19T16:39:59,A,17,C",
            "1996-12-19T16:40:00,A,9,C",
            "1996-12-19T16:40:01,A,12,C",
            "1996-12-19T16:40:02,A,,C",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_with_key_literal(source, golden) -> None:
    golden.jsonl(source.with_key("literal_key"))


async def test_with_key_column(source, golden) -> None:
    new_key = source.col("new_key")
    golden.jsonl(source.with_key(new_key))


async def test_with_key_grouping(source, golden) -> None:
    new_key = source.col("new_key")
    grouping = "user"
    golden.jsonl(source.with_key(new_key, grouping))


async def test_with_key_last(source, golden) -> None:
    golden.jsonl(source.with_key(source.col("new_key")).last())



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m",
            "1996-12-19T16:39:57,A,5",
            "1996-12-19T16:39:58,B,100.0001",
            "1996-12-19T16:39:59,A,2.50",
            "1996-12-19T16:40:00,A,",
            "1996-12-19T16:40:01,A,0.99",
            "1996-12-19T16:40:02,A,1.01",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_clamp_min_max(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(kd.record({"m": m, "clamped_m": m.clamp(min=5, max=100)}))


async def test_clamp_min(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(kd.record({"m": m, "clamped_min": m.clamp(min=5)}))


async def test_clamp_max(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(kd.record({"m": m, "clamped_min": m.clamp(max=100)}))


@pytest.fixture(scope="module")
async def banking_source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,current_balance,min_balance,max_balance",
            "1996-12-19T16:39:57,A,5.00,6.00,6.01",
            "1996-12-19T16:39:58,B,6.00,7.00,7.01",
            "1996-12-19T16:39:59,A,7.00,8.00,8.01",
            "1996-12-19T16:40:00,A,8.00,9.00,9.01",
            "1996-12-19T16:40:01,A,9.00,10.00,10.01",
            "1996-12-19T16:40:02,A,10.00,11.00,11.01",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_clamp_banking_min_max(banking_source, golden) -> None:
    current_balance = banking_source.col("current_balance")
    min_balance = banking_source.col("min_balance")
    max_balance = banking_source.col("max_balance")
    golden.jsonl(
        kd.record(
            {
                "current_balance": current_balance,
                "clamped_balance": current_balance.clamp(
                    min=min_balance, max=max_balance
                ),
            }
        )
    )


async def test_clamp_banking_min(banking_source, golden) -> None:
    current_balance = banking_source.col("current_balance")
    min_balance = banking_source.col("min_balance")
    golden.jsonl(
        kd.record(
            {
                "current_balance": current_balance,
                "clamped_balance": current_balance.clamp(min=min_balance),
            }
        )
    )


async def test_clamp_banking_max(banking_source, golden) -> None:
    current_balance = banking_source.col("current_balance")
    max_balance = banking_source.col("max_balance")
    golden.jsonl(
        kd.record(
            {
                "current_balance": current_balance,
                "clamped_balance": current_balance.clamp(max=max_balance),
            }
        )
    )



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,a,b",
            "2021-01-01T00:00:00,A,5.7,1.2",
            "2021-01-02T00:00:00,A,6.3,0.4",
            "2021-01-03T00:00:00,B,,3.7",
            "2021-01-04T00:00:00,A,13.2,",
            "2021-01-05T00:00:00,A,2,5.4",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_least(source, golden) -> None:
    a = source.col("a")
    b = source.col("b")
    golden.jsonl(kd.record({"a": a, "b": b, "a_least_b": a.least(b)}))



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m",
            "1996-12-19T16:39:57,A,1",
            "1996-12-19T16:39:58,B,1",
            "1996-12-19T16:39:59,A,2",
            "1996-12-19T16:40:00,A,3",
            "1996-12-19T16:40:01,A,4",
            "1996-12-19T16:40:02,A,5",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_exp(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(kd.record({"m": m, "exp_m": m.exp()}))



"""Nox sessions."""
import os
import shutil
import sys
from pathlib import Path
from typing import Iterable
from typing import Iterator

import nox

package = "kaskada"
python_versions = ["3.11", "3.10", "3.9"]
nox.needs_version = ">= 2021.6.6"
nox.options.sessions = (
    "check-lint",
    "safety",
    "mypy",
    "tests",
    "typeguard",
    "xdoctest",
    "docs-build",
)


@nox.session(name="check-lint", python=python_versions[0])
def check_lint(session: nox.Session) -> None:
    """Lint."""
    args = session.posargs or ["pysrc", "pytests", "docs/source"]
    install(session, groups=["lint"], root=False)
    session.run("black", "--check", *args)
    session.run("flake8", *args)
    session.run("isort", "--filter-files", "--check-only", *args)

    session.run("pydocstyle", "--convention=google", "pysrc")
    # No way to run this as a check.
    # session.run("pyupgrade", "--py38-plus")


@nox.session(name="fix-lint", python=python_versions[0])
def fix_lint(session: nox.Session) -> None:
    """Automatically fix lint issues."""
    args = session.posargs or ["pysrc", "pytests", "docs/source"]
    install(session, groups=["lint"], root=False)
    session.run("autoflake", "--in-place", "--remove-all-unused-imports", "--recursive", *args)
    session.run("isort", "--filter-files", *args)
    session.run("pyupgrade", "--py38-plus")
    session.run("black", *args)


@nox.session(python=python_versions[0])
def safety(session: nox.Session) -> None:
    """Scan dependencies for insecure packages."""
     # NOTE: Pass `extras` to `export_requirements` if the project supports any.
    requirements = export_requirements(session)
    install(session, groups=["safety"], root=False)
    session.run("safety", "check", "--full-report", f"--file={requirements}")


@nox.session(python=python_versions)
def mypy(session: nox.Session) -> None:
    """Type-check using mypy."""
    args = session.posargs or ["pysrc", "pytests"]
    install(session, groups=["typecheck"])
    # Using `--install-types` should make this less picky about missing stubs.
    # However, there is a possibility it slows things down, by making mypy
    # run twice -- once to determine what types need to be installed, then once
    # to check things with those stubs.
    session.run("mypy", "--install-types", "--non-interactive", *args)
    if not session.posargs:
        session.run("mypy", f"--python-executable={sys.executable}", "noxfile.py")


@nox.session(python=python_versions)
def tests(session: nox.Session) -> None:
    """Run the test suite."""
    install(session, groups=["test"])
    try:
        session.run("coverage", "run", "--parallel", "-m", "pytest", *session.posargs)
    finally:
        if session.interactive:
            session.notify("coverage", posargs=[])


@nox.session(python=python_versions[0])
def coverage(session: nox.Session) -> None:
    """Produce the coverage report."""
    args = session.posargs or ["report"]

    install(session, groups=["test"])

    if not session.posargs and any(Path().glob(".coverage.*")):
        session.run("coverage", "combine")

    session.run("coverage", *args)


@nox.session(python=python_versions[0])
def typeguard(session: nox.Session) -> None:
    """Runtime type checking using Typeguard."""
    install(session, groups=["typecheck", "test"])
    session.run("pytest", f"--typeguard-packages={package}", *session.posargs)


@nox.session(python=python_versions)
def xdoctest(session: nox.Session) -> None:
    """Run examples with xdoctest."""
    if session.posargs:
        args = [package, *session.posargs]
    else:
        args = [f"--modname={package}", "--command=all"]
        if "FORCE_COLOR" in os.environ:
            args.append("--colored=1")

    install(session, groups=["test"])
    session.run("python", "-m", "xdoctest", *args)

@nox.session(name="docs-build", python=python_versions[0])
def docs_build(session: nox.Session) -> None:
    """Build the documentation."""
    # ablog doesn't currently indicate whether it supports parallel reads,
    # leading to a warning.
    # when possible, add `"-j", "auto",` to do parallel builds (and in CI).
    args = session.posargs or ["docs/source", "docs/_build", "-W"]
    if not session.posargs and "FORCE_COLOR" in os.environ:
        args.insert(0, "--color")

    install(session, groups=["typecheck", "docs"])

    build_dir = Path("docs", "_build")
    if build_dir.exists():
        shutil.rmtree(build_dir)

    session.run("sphinx-build", *args)


@nox.session(python=python_versions[0])
def docs(session: nox.Session) -> None:
    """Build and serve the documentation with live reloading on file changes."""
    args = ["--open-browser", "docs/source", "docs/_build", "-j", "auto", "--ignore", "*/apidocs/*", "--watch", "pysrc/kaskada"]
    install(session, groups=["typecheck", "docs"])

    build_dir = Path("docs", "_build")
    if build_dir.exists():
        shutil.rmtree(build_dir)

    session.run("sphinx-autobuild", *args)


def install(session: nox.Session, *, groups: Iterable[str], root: bool = True) -> None:
    """Install the dependency groups using Poetry.
    This function installs the given dependency groups into the session's
    virtual environment. When ``root`` is true (the default), the function
    also installs the root package's default dependencies.

    The root package is installed using `maturin develop`.

    Args:
        session: The Session object.
        groups: The dependency groups to install.
        root: Install the root package.
    """
    session.run_always(
        "poetry",
        "install",
        "--no-root",
        "--sync",
        "--{}={}".format("only" if not root else "with", ",".join(groups)),
        external=True,
    )
    if root:
        session.run_always("maturin", "develop", "--profile", "dev")


def export_requirements(session: nox.Session, *, extras: Iterable[str] = ()) -> Path:
    """Export a requirements file from Poetry.
    This function uses ``poetry export`` to generate a requirements file
    containing the default dependencies at the versions specified in
    ``poetry.lock``.

    Args:
        session: The Session object.
        extras: Extras supported by the project.
    Returns:
        The path to the requirements file.
    """
    # XXX Use poetry-export-plugin with dependency groups
    output = session.run_always(
        "poetry",
        "export",
        "--format=requirements.txt",
        "--without-hashes",
        *[f"--extras={extra}" for extra in extras],
        external=True,
        silent=True,
        stderr=None,
    )

    if output is None:
        session.skip(
            "The command `poetry export` was not executed"
            " (a possible cause is specifying `--no-install`)"
        )

    assert isinstance(output, str)  # noqa: S101

    def _stripwarnings(lines: Iterable[str]) -> Iterator[str]:
        for line in lines:
            if line.startswith("Warning:"):
                print(line, file=sys.stderr)
                continue
            yield line

    text = "".join(_stripwarnings(output.splitlines(keepends=True)))

    path = session.cache_dir / "requirements.txt"
    path.write_text(text)

    return path



from typing import Callable, List, Literal, Optional, Sequence

import pyarrow as pa

from ._execution import _ExecutionOptions

class Session:
    def __init__(self) -> None: ...

class Execution(object):
    def schema(self) -> pa.Schema: ...
    def collect_pyarrow(self) -> List[pa.RecordBatch]: ...
    def next_pyarrow(self) -> Optional[pa.RecordBatch]: ...
    def stop(self) -> None: ...
    async def next_pyarrow_async(self) -> Optional[pa.RecordBatch]: ...

class Expr:
    @staticmethod
    def call(
        session: Session,
        operation: str,
        args: Sequence[Expr],
    ) -> Expr: ...
    @staticmethod
    def call_udf(
        session: Session,
        udf: Udf,
        args: Sequence[Expr],
    ) -> Expr: ...
    @staticmethod
    def literal(
        session: Session,
        value: int | float | str | None,
    ) -> Expr: ...
    @staticmethod
    def literal_timedelta(
        session: Session,
        s: int,
        ns: int,
    ) -> Expr: ...
    def cast(self, data_type: pa.DataType) -> Expr: ...
    def data_type(self) -> pa.DataType: ...
    def is_continuous(self) -> bool: ...
    def session(self) -> Session: ...
    def execute(self, options: Optional[_ExecutionOptions] = None) -> Execution: ...
    def grouping(self) -> Optional[str]: ...
    def explain(
        self,
        kind: Literal["initial_dfg", "final_dfg", "final_plan"],
        options: Optional[_ExecutionOptions],
    ) -> str: ...

class Table:
    def __init__(
        self,
        session: Session,
        name: str,
        time_column: str,
        key_column: str,
        schema: pa.Schema,
        queryable: bool,
        subsort_column: Optional[str],
        grouping_name: Optional[str],
        time_unit: Optional[str],
        source: Optional[str],
    ) -> None: ...
    @property
    def name(self) -> str: ...
    def expr(self) -> Expr: ...
    async def add_pyarrow(self, data: pa.RecordBatch) -> None: ...
    async def add_parquet(self, file: str) -> None: ...

class Udf(object):
    def __init__(self, result_ty: str, result_fn: Callable[..., pa.Array]) -> None: ...

async def parquet_schema(session: Session, url: str) -> pa.Schema: ...



"""Sphinx configuration."""
import sys
from pathlib import Path
from typing import Any, Dict


sys.path.append(str(Path(".").resolve()))

project = "kaskada"
author = "Kaskada Contributors"
copyright = "2023, Kaskada Contributors"
extensions = [
    "ablog",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx_design",
    # "myst_parser",
    "myst_nb",
    "sphinx_copybutton",
    "sphinx_autodoc_typehints",  # must be after napoleon
    "sphinx_social_cards",
    "_extensions.gallery_directive",
]
autodoc_typehints = "description"
language = "en"

html_theme = "pydata_sphinx_theme"
html_favicon = "_static/favicon.png"
html_title = "Kaskada"
html_js_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js"
]

# Configure the primary (left) sidebar.
# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/layout.html#primary-sidebar-left
html_sidebars = {
    # No primary (left) sidebar on the landing page.
    "index.md": [],
    # Blog sidebars
    # https://ablog.readthedocs.io/manual/ablog-configuration-options/#blog-sidebars
    "blog/**": [
        # Information about the post.
        "ablog/postcard.html",
        # 5 most recent posts
        "ablog/recentposts.html",
        # Tag cloud and links.
        "ablog/tagcloud.html",
        # Categories -- we just use tags for now.
        # "ablog/categories.html",
        # Show all authors on the sidebar.
        # "ablog/authors.html",
        # Show all languages on the sidebar.
        # "ablog/languages.html",
        # Show all locations on the sidebar.
        # "ablog/locations.html",
        "ablog/archives.html",
    ],
    "[!blog]*/*": ["sidebar-nav-bs"],
}

html_theme_options: Dict[str, Any] = {
    # Setup external links.
    # https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/header-links.html
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/kaskada-ai/kaskada",  # required
            "icon": "fa-brands fa-square-github",
            "type": "fontawesome",
        },
        {
            "name": "Slack",
            "url": "https://join.slack.com/t/kaskada-hq/shared_invite/zt-1t1lms085-bqs2jtGO2TYr9kuuam~c9w",
            "icon": "fa-brands fa-slack",
        },
    ],
    # Setup edit buttons
    # See https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/source-buttons.html
    "use_edit_page_button": True,
    # Include indices.
    # https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/indices.html
    "primary_sidebar_end": ["indices.html"],
    # Provide an announcement at the top.
    # https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/announcements.html#
    "announcement": (
        "This describes the next version of Kaskada. "
        "It is currently available as an alpha release."
    ),
    # Branding / logos.
    # https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/branding.html
    "logo": {
        "image_light": "_static/kaskada-positive.svg",
        "image_dark": "_static/kaskada-negative.svg",
    },
    # Setup analytics.
    # https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/analytics.html
    "analytics": {
        "google_analytics_id": "G-HR9E2E6TG4",
    },
    # TODO: Version switcher.
    # This would require hosting multiple versions of the docs.
    # https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/version-dropdown.html
}

templates_path = ["_templates"]
html_static_path = ["_static"]

html_context = {
    "github_user": "kaskada-ai",
    "github_repo": "kaskada",
    "github_version": "main",
    "doc_path": "python/docs/source",
}

# Setup links to other sphinx projects.
intersphinx_mapping: Dict[str, Any] = {
    "python": ("http://docs.python.org/3", None),
    "pandas": ("http://pandas.pydata.org/docs", None),
    "pyarrow": ("https://arrow.apache.org/docs", None),
}

# adds useful copy functionality to all the examples; also
# strips the '>>>' and '...' prompt/continuation prefixes.
copybutton_prompt_text = r">>> |\.\.\. "
copybutton_prompt_is_regexp = True

# Options for Todos
todo_include_todos = True

# Options for Myst (markdown)
# https://myst-parser.readthedocs.io/en/v0.17.1/syntax/optional.html
myst_enable_extensions = [
    "colon_fence",
    "deflist",
    "smartquotes",
    "replacements",
    "linkify",
]
myst_heading_anchors = 3

# -- Options for autodoc ----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

# Automatically extract typehints when specified and place them in
# descriptions of the relevant function/method.
autodoc_typehints = "description"
autodoc_type_aliases = {"kaskada.Arg": "kaskada.Arg"}

# Don't show class signature with the class' name.
autodoc_class_signature = "separated"

autosummary_generate = True

napoleon_preprocess_types = True
napoleon_attr_annotations = True
napoleon_use_rtype = False
typehints_use_rtype = False
typehints_document_rtype = False
typehints_defaults = "comma"

suppress_warnings = ["mystnb.unknown_mime_type"]

blog_path = "blog"
blog_authors = {
    "ben": ("Ben Chambers", "https://github.com/bjchambers"),
    "ryan": ("Ryan Michael", "https://github.com/kerinin"),
}
post_date_format = "%Y-%b-%d"
post_date_format_short = "%Y-%b-%d"
post_show_prev_next = False

# Generate social cards for blog posts
social_cards = {
    "site_url": "https://kaskada.io/kaskada",
    "description": "Kaskada: Real-Time AI without the fuss.",
    "cards_layout_dir": ["_layouts"],
    "cards_layout_options": {
        "background_color": "#26364a",
    },
}



import datetime as datetime

import kaskada as kd
import pyarrow as pa
import pytest


@pytest.fixture
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n,t",
            "1996-12-19T16:39:57,A,5,10,1996-12-19T16:42:57",
            "1996-12-19T16:39:58,B,24,3,1996-12-19T16:39:59",
            "1996-12-19T16:39:59,A,17,6,",
            "1996-12-19T16:40:00,A,,9,1996-12-19T16:41:00",
            "1996-12-19T16:40:01,A,12,,1996-12-19T16:42:01",
            "1996-12-19T16:40:02,A,,,1996-12-19T16:43:02",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_seconds_since(golden, source) -> None:
    t1 = source.col("time")
    t2 = source.col("t")
    golden.jsonl(
        kd.record(
            {
                "t1": t1,
                "t2": t2,
                "seconds_since_t1": t2.seconds_since(t1).cast(pa.int64()),
                "seconds_since_t2": t1.seconds_since(t2).cast(pa.int64()),
            }
        )
    )


async def test_seconds_since_datetime(golden, source) -> None:
    t = source.col("time")
    dt = datetime.datetime(1996, 12, 19, 16, 39, 50, tzinfo=datetime.timezone.utc)
    golden.jsonl(
        kd.record(
            {"t1": t, "seconds_since_literal": t.seconds_since(dt).cast(pa.int64())}
        )
    )



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n,default",
            "1996-12-19T16:39:57,A,5,10,-1",
            "1996-12-19T16:39:58,B,24,3,-1",
            "1996-12-19T16:39:59,A,17,6,-1",
            "1996-12-19T16:40:00,A,,9,-1",
            "1996-12-19T16:40:01,A,12,,-1",
            "1996-12-19T16:40:02,A,,,-1",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_else_(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    default = source.col("default")
    condition_m = m > 15
    condition_n = n > 5
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "condition_m": condition_m,
                "if_else_m": m.if_(condition_m).else_(default),
                "n": n,
                "condition_n": condition_n,
                "if_n": n.if_(condition_n).else_(default),
            }
        )
    )


@pytest.fixture(scope="module")
async def record_source() -> kd.sources.JsonlString:
    content = "\n".join(
        [
            """{"time":"1996-12-19T16:39:57","key":"A","override": {"test":"override_val"}}""",
            """{"time":"1996-12-19T16:39:58","key":"A","default_record":{"test":"default"}}""",
        ]
    )
    return await kd.sources.JsonlString.create(
        content, time_column="time", key_column="key"
    )


async def test_else_debug(record_source, golden) -> None:
    default_record = record_source.col("default_record")
    override_column = record_source.col("override")
    golden.jsonl(
        kd.record(
            {
                "default_record": default_record,
                "overide": override_column,
                "override_else_default": override_column.else_(default_record),
            }
        )
    )



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1996-12-19T16:40:00,A,,9",
            "1996-12-19T16:40:01,A,12,",
            "1996-12-19T16:40:02,A,,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_variance_unwindowed(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(
        kd.record(
            {"m": m, "variance_m": m.variance(), "n": n, "variance_n": n.variance()}
        )
    )


async def test_variance_windowed(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "since_variance_m": m.variance(window=kd.windows.Since(m > 20)),
                "n": n,
                "sliding_variance_n": n.variance(window=kd.windows.Sliding(2, m > 10)),
            }
        )
    )


async def test_variance_since_true(source, golden) -> None:
    # `since(True)` should be the same as unwindowed, so equals the original vaule.
    m_variance_since_true = kd.record(
        {
            "m": source.col("m"),
            "since_m_variance": source.col("m").variance(window=kd.windows.Since(True)),
        }
    )
    golden.jsonl(m_variance_since_true)



"""Defines classes representing Kaskada expressions."""

from __future__ import annotations

import sys
import warnings
from datetime import datetime, timedelta
from typing import (
    TYPE_CHECKING,
    Callable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    final,
    overload,
)

import kaskada as kd
import kaskada._ffi as _ffi
import pandas as pd
import pyarrow as pa
from typing_extensions import TypeAlias

from ._execution import Execution, ResultIterator, _ExecutionOptions


if TYPE_CHECKING:
    import graphviz

#: A literal value that can be used as an argument to a Timestream operation.
LiteralValue: TypeAlias = Optional[Union[int, str, float, bool, timedelta, datetime]]

#: A Timestream or literal which can be used as an argument to a Timestream operation.
Arg: TypeAlias = Union[
    "Timestream", Callable[["Timestream"], "Timestream"], LiteralValue
]


def _augment_error(args: Sequence[Arg], e: Exception) -> Exception:
    """Augment an error with information about the arguments."""
    if sys.version_info >= (3, 11):
        # If we can add notes to the exception, indicate the types.
        # This works in Python >=3.11
        for n, arg in enumerate(args):
            if isinstance(arg, Timestream):
                e.add_note(f"Arg[{n}]: Timestream[{arg.data_type}]")
            else:
                e.add_note(f"Arg[{n}]: Literal {arg} ({type(arg)})")
    return e


def _extract_arg(
    arg: Arg, input: Optional[Timestream], session: _ffi.Session
) -> _ffi.Expr:
    """Extract the FFI expression from an argument."""
    if callable(arg):
        if input is None:
            raise ValueError(
                "Cannot use a callable argument without an input Timestream"
            )
        else:
            arg = arg(input)

    if isinstance(arg, Timestream):
        return arg._ffi_expr
    else:
        return Timestream._literal(arg, session)._ffi_expr


class Timestream(object):
    """A `Timestream` represents a computation producing a Timestream."""

    _ffi_expr: _ffi.Expr

    def __init__(self, ffi: _ffi.Expr) -> None:
        """Create a new expression."""
        self._ffi_expr = ffi

    @staticmethod
    def _literal(value: LiteralValue, session: _ffi.Session) -> Timestream:
        """Construct a Timestream for a literal value."""
        if isinstance(value, datetime):
            raise TypeError("Cannot create a literal Timestream from a datetime")

        if isinstance(value, timedelta):
            # The smallest unit stored in timedelta is us
            us = value / timedelta(microseconds=1)
            # Get the seconds as an integer
            seconds = int(us / 1_000_000)
            # Get the leftover nanoseconds
            nanoseconds = int((us % 1_000_000) * 1_000)
            return Timestream(
                _ffi.Expr.literal_timedelta(session, seconds, nanoseconds)
            )
        else:
            return Timestream(_ffi.Expr.literal(session, value))

    @staticmethod
    def _call(
        func: Union[str, _ffi.Udf],
        *args: Arg,
        input: Optional[Timestream] = None,
        session: Optional[_ffi.Session] = None,
    ) -> Timestream:
        """Construct a new Timestream by calling the given function.

        Args:
            func: Name of the function to apply.
            input: The input to use for any "deferred" arguments.
              If `None`, then any arguments that require a `Timestream` argument
              will produce an error.
            *args: List of arguments to the expression.
            session: FFI Session to create the expression in.
              If unspecified, will infer from the arguments.
              Will fail if all arguments are literals and the session is not provided.

        Returns:
            Timestream representing the result of the function applied to the arguments.

        Raises:
            # noqa: DAR401 _augment_error
            TypeError: If the argument types are invalid for the given function.
            ValueError: If the argument values are invalid for the given function.
        """
        # Determine the session. This is unfortunately necessary for creating literals
        # and other expressions.
        if session is None:
            if input is None:
                # If there is no input, then at least one of the arguments must be a
                # Timestream. Specifically, we shouldn't have all literals, nor should
                # we have all Callables that produce Timestreams.
                try:
                    session = next(
                        arg._ffi_expr.session()
                        for arg in args
                        if isinstance(arg, Timestream)
                    )
                except StopIteration as e:
                    raise ValueError(
                        "Cannot determine session from only literal arguments. "
                        "Please provide a session explicitly or use at least one non-literal."
                    ) from e
            else:
                # If there is a session, it is possible (but unlikely) that all arguments
                # are Callables. To handle this, we need to determine the session from the
                # input.
                session = input._ffi_expr.session()

        ffi_args = [_extract_arg(arg, input=input, session=session) for arg in args]
        try:
            if isinstance(func, str):
                return Timestream(
                    _ffi.Expr.call(session=session, operation=func, args=ffi_args)
                )
            elif isinstance(func, _ffi.Udf):
                return Timestream(
                    _ffi.Expr.call_udf(session=session, udf=func, args=ffi_args)
                )
            else:
                raise TypeError(
                    f"invalid type for func. Expected str or udf, saw: {type(func)}"
                )
        except TypeError as e:
            # noqa: DAR401
            raise _augment_error(args, TypeError(str(e))) from e
        except ValueError as e:
            raise _augment_error(args, ValueError(str(e))) from e

    @property
    def data_type(self) -> pa.DataType:
        """The PyArrow type of values in this Timestream."""
        return self._ffi_expr.data_type()

    @property
    def is_continuous(self) -> bool:
        """Returns true if this Timestream is continuous."""
        return self._ffi_expr.is_continuous()

    @final
    def pipe(
        self,
        func: Union[Callable[..., Timestream], Tuple[Callable[..., Timestream], str]],
        *args: Arg,
        **kwargs: Arg,
    ) -> Timestream:
        """Apply chainable functions that produce Timestreams.

        Args:
            func: Function to apply to this Timestream.
              Alternatively a `(func, keyword)` tuple where `keyword` is a string
              indicating the keyword of `func` that expects the Timestream.
            *args: Positional arguments passed into ``func``.
            **kwargs: A dictionary of keyword arguments passed into ``func``.

        Returns:
            The result of applying `func` to the arguments.

        Raises:
            ValueError: When using `self` with a specific `keyword` if the `keyword` also
              appears on in the `kwargs`.

        Notes:
            Use ``.pipe`` when chaining together functions that expect Timestreams.

        Examples:
            Instead of writing

            >>> func(g(h(df), arg1=a), arg2=b, arg3=c)  # doctest: +SKIP

            You can write

            >>> (df.pipe(h)
            >>>    .pipe(g, arg1=a)
            >>>    .pipe(func, arg2=b, arg3=c)
            >>> )  # doctest: +SKIP

            If you have a function that takes the data as (say) the second
            argument, pass a tuple indicating which keyword expects the
            data. For example, suppose ``func`` takes its data as ``arg2``:

            >>> (df.pipe(h)
            >>>    .pipe(g, arg1=a)
            >>>    .pipe((func, 'arg2'), arg1=a, arg3=c)
            >>>  )  # doctest: +SKIP
        """
        if isinstance(func, tuple):
            func, target = func
            if target in kwargs:
                msg = f"{target} is both the pipe target and a keyword argument"
                raise ValueError(msg)
            kwargs[target] = self
            return func(*args, **kwargs)
        else:
            return func(self, *args, **kwargs)

    def add(self, rhs: Arg) -> Timestream:
        """Return a Timestream adding this and `rhs`.

        Args:
            rhs: The Timestream or literal value to add to this.

        Notes:
            You can also write `a.add(b)` as `a + b`.
        """
        if isinstance(rhs, timedelta):
            # Right now, we can't convert a time delta directly to a scalar value (literal).
            # So we convert it to seconds and then add it.
            # Note that this loses precision if the timedelta has a fractional number of seconds,
            # and fail if the number of seconds exceeds an integer.
            session = self._ffi_expr.session()
            seconds = Timestream._call(
                "seconds", int(rhs.total_seconds()), session=session
            )
            return Timestream._call("add_time", seconds, self, input=self)
        else:
            return Timestream._call("add", self, rhs, input=self)

    def __add__(self, rhs: Arg) -> Timestream:
        """Implement `self + rhs`."""
        return self.add(rhs)

    def __radd__(self, lhs: Arg) -> Timestream:
        """Implement `lhs + self`."""
        if callable(lhs):
            lhs = lhs(self)
        if not isinstance(lhs, Timestream):
            lhs = Timestream._literal(lhs, self._ffi_expr.session())
        return lhs.add(self)

    def ceil(self) -> Timestream:
        """Return a Timestream rounding self up to the next largest integer."""
        return Timestream._call("ceil", self, input=self)

    def clamp(self, min: Arg = None, max: Arg = None) -> Timestream:
        """Return a Timestream from `self` clamped between `min` and `max`.

        Args:
            min: The literal value to set as the lower bound.
            max: The literal value to set as the upper bound.
        """
        return Timestream._call("clamp", self, min, max, input=self)

    def sub(self, rhs: Arg) -> Timestream:
        """Return a Timestream subtracting `rhs` from this.

        Args:
            rhs: The Timestream or literal value to subtract from this.

        Notes:
            You can also write `a.sub(b)` as `a - b`.
        """
        return Timestream._call("sub", self, rhs, input=self)

    def __sub__(self, rhs: Arg) -> Timestream:
        """Implement `self - rhs`."""
        return self.sub(rhs)

    def __rsub__(self, lhs: Arg) -> Timestream:
        """Implement `lhs - self`."""
        if callable(lhs):
            lhs = lhs(self)
        if not isinstance(lhs, Timestream):
            lhs = Timestream._literal(lhs, self._ffi_expr.session())
        return lhs.sub(self)

    def exp(self) -> Timestream:
        """Return a Timestream raising `e` to the power of `self`."""
        return Timestream._call("exp", self)

    def floor(self) -> Timestream:
        """Return a Timestream of self rounded down to the nearest integer."""
        return Timestream._call("floor", self)

    def hash(self) -> Timestream:
        """Return a Timestream containing the hash of the input.

        Notes:
            This will only return `null` when interpolated at points where
            the input is not defined. At other points, it will return the
            hash of `null`, which is `0` (not `null`).
        """
        return Timestream._call("hash", self)

    def len(self) -> Timestream:
        """Return a Timestream with the length of input string.

        Notes:
            Returns `0` for an empty string and `null` for `null`
        """
        return Timestream._call("len", self)

    def lower(self) -> Timestream:
        """Return a Timestream with all values converted to lower case."""
        return Timestream._call("lower", self)

    def mul(self, rhs: Arg) -> Timestream:
        """Return a Timestream multiplying this and `rhs`.

        Args:
            rhs: The Timestream or literal value to multiply with this.

        Notes:
            You can also write `a.mul(b)` as `a * b`.
        """
        return Timestream._call("mul", self, rhs, input=self)

    def __mul__(self, rhs: Arg) -> Timestream:
        """Implement `self * rhs`."""
        return self.mul(rhs)

    def __rmul__(self, lhs: Arg) -> Timestream:
        """Implement `lhs * self`."""
        if callable(lhs):
            lhs = lhs(self)
        if not isinstance(lhs, Timestream):
            lhs = Timestream._literal(lhs, self._ffi_expr.session())
        return lhs.mul(self)

    def powf(self, power: Arg) -> Timestream:
        """Return a Timestream raising `self` to the power of `power`.

        Args:
            power: The Timestream or literal value to raise `self` to.
        """
        return Timestream._call("powf", self, power, input=self)

    def div(self, divisor: Arg) -> Timestream:
        """Return a Timestream by dividing this and `divisor`.

        Args:
            divisor: The Timestream or literal value to divide this by.

        Notes:
            You can also write `a.div(b)` as `a / b`.
        """
        return Timestream._call("div", self, divisor, input=self)

    def __truediv__(self, divisor: Arg) -> Timestream:
        """Implement `self / divisor`."""
        return self.div(divisor)

    def __rtruediv__(self, dividend: Arg) -> Timestream:
        """Implement `dividend / self`."""
        if callable(dividend):
            dividend = dividend(self)
        if not isinstance(dividend, Timestream):
            dividend = Timestream._literal(dividend, self._ffi_expr.session())
        return dividend.div(self)

    def lt(self, rhs: Arg) -> Timestream:
        """Return a Timestream that is true if this is less than `rhs`.

        Args:
            rhs: The Timestream or literal value to compare to.

        Notes:
            You can also write `a.lt(b)` as `a < b`.
        """
        return Timestream._call("lt", self, rhs, input=self)

    def __lt__(self, rhs: Arg) -> Timestream:
        """Implement `self < rhs`."""
        return self.lt(rhs)

    def le(self, rhs: Arg) -> Timestream:
        """Return a Timestream that is true if this is less than or equal to `rhs`.

        Args:
            rhs: The Timestream or literal value to compare to.

        Notes:
            You can also write `a.le(b)` as `a <= b`.
        """
        return Timestream._call("lte", self, rhs, input=self)

    def __le__(self, rhs: Arg) -> Timestream:
        """Implement `self <= rhs`."""
        return self.le(rhs)

    def gt(self, rhs: Arg) -> Timestream:
        """Return a Timestream that is true if this is greater than `rhs`.

        Args:
            rhs: The Timestream or literal value to compare to.

        Notes:
            You can also write `a.gt(b)` as `a > b`.
        """
        return Timestream._call("gt", self, rhs, input=self)

    def __gt__(self, rhs: Arg) -> Timestream:
        """Implement `self > rhs`."""
        return self.gt(rhs)

    def ge(self, rhs: Arg) -> Timestream:
        """Return a TimeStream that is true if this is greater than or equal to `rhs`.

        Args:
            rhs: The Timestream or literal value to compare to.

        Notes:
            You can also write `a.ge(b)` as `a >= b`.
        """
        return Timestream._call("gte", self, rhs, input=self)

    def __ge__(self, rhs: Arg) -> Timestream:
        """Implement `self >= rhs`."""
        return self.ge(rhs)

    def and_(self, rhs: Arg) -> Timestream:
        """Return the logical conjunction of this Timestream and `rhs`.

        Args:
            rhs: The Timestream or literal value to conjoin with.
        """
        return Timestream._call("logical_and", self, rhs, input=self)

    def or_(self, rhs: Arg) -> Timestream:
        """Return the logical disjunction of this Timestream and `rhs`.

        Args:
            rhs: The Timestream or literal value to disjoin with.
        """
        return Timestream._call("logical_or", self, rhs, input=self)

    def not_(self) -> Timestream:
        """Return the logical negation of this Timestream."""
        return Timestream._call("not", self)

    def eq(self, other: Arg) -> Timestream:
        """Return a Timestream that is true if this is equal to `other`.

        Args:
            other: The Timestream or literal value to compare to.

        Note:
            Equality is *not* available as `a == b`.
        """
        return Timestream._call("eq", self, other, input=self)

    def ne(self, other: Arg) -> Timestream:
        """Return a Timestream that is true if this is not equal to `other`.

        Args:
            other: The Timestream or literal value to compare to.

        Note:
            Inequality is *not* available as `a != b`.
        """
        return Timestream._call("neq", self, other, input=self)

    def __eq__(self, other: object) -> bool:
        """Warn when Timestreams are compared using `==`."""
        warnings.warn(
            "Using '==' with Timestreams doesn't produce a boolean stream. Use 'eq' instead.",
            stacklevel=2,
        )
        return super().__eq__(other)

    def __ne__(self, other: object) -> bool:
        """Warn when Timestreams are compared using `!=`."""
        warnings.warn(
            "Using '!=' with Timestreams doesn't produce a boolean stream. Use 'ne' instead.",
            stacklevel=2,
        )
        return super().__ne__(other)

    def index(self, key: Arg) -> Timestream:
        """Return a Timestream indexing into the elements of `self`.

        If the Timestream contains lists, the key should be an integer index.

        If the Timestream contains maps, the key should be the same type as the map keys.

        Args:
            key: The key to index into the expression.

        Raises:
            TypeError: When the Timestream is not a record, list, or map.

        Returns:
            Timestream with the resulting value (or `null` if absent) at each point.

        Note:
            Indexing may be written using the operator `self[key]` instead of `self.index(key)`.
        """
        data_type = self.data_type
        if isinstance(data_type, pa.MapType):
            return Timestream._call("get", key, self, input=self)
        elif isinstance(data_type, pa.ListType):
            return Timestream._call("index", key, self, input=self)
        else:
            raise TypeError(f"Cannot index into {data_type}")

    def __getitem__(self, key: Arg) -> Timestream:
        """Implement `self[key]` using `index`.

        See Also:
            index
        """
        return self.index(key)

    def col(self, name: str) -> Timestream:
        """Return a Timestream accessing the named column or field of `self`.

        Args:
            name: The name of the column or field to access.

        Raises:
            TypeError: When the Timestream is not a record.
        """
        data_type = self.data_type
        if isinstance(data_type, pa.StructType) or isinstance(data_type, pa.ListType):
            return Timestream._call("fieldref", self, name)
        else:
            raise TypeError(
                f"Cannot access column {name!r} of non-record type '{data_type}'"  # noqa : B907
            )

    def select(self, *args: str) -> Timestream:
        """Return a Timestream selecting the given fields from `self`.

        Args:
            *args: The field names to select.
        """
        return Timestream._call("select_fields", self, *args)

    def substring(
        self, start: Optional[int] = None, end: Optional[int] = None
    ) -> Timestream:
        """Return a Timestream with a substring between the start and end indices.

        Args:
            start: The inclusive index to start at. `None` indicates the beginning
                of the string. Negative indices count backwards from the end of
                the string.
            end: The exclusive index to end at. `None` indicates the length of
                the string. Negative indices count backwards from the end of
                the string.

        Notes:
            Returns the substring starting at `start` (inclusive) up to but not
            including the `end`.

            If the input is `null`, returns `null`. If `end` > `start` an empty
            string is returned.
        """
        return Timestream._call("substring", self, start, end)

    def remove(self, *args: str) -> Timestream:
        """Return a Timestream removing the given fields from `self`.

        Args:
            *args: The field names to remove.
        """
        return Timestream._call("remove_fields", self, *args)

    def extend(
        self,
        fields: Timestream
        | Mapping[str, Arg]
        | Callable[[Timestream], Timestream | Mapping[str, Arg]],
    ) -> Timestream:
        """Return a Timestream containing fields from `self` and `fields`.

        If a field exists in the base Timestream and the `fields`, the value
        from the `fields` will be taken.

        Args:
            fields: Fields to add to each record in the Timestream.
        """
        # This argument order is weird, and we shouldn't need to make a record
        # in order to do the extension.
        if callable(fields):
            fields = fields(self)
        if not isinstance(fields, Timestream):
            fields = record(fields)
        return Timestream._call("extend_record", fields, self, input=self)

    def neg(self) -> Timestream:
        """Return a Timestream from the numeric negation of self."""
        return Timestream._call("neg", self)

    def is_null(self) -> Timestream:
        """Return a boolean Timestream containing `true` when `self` is `null`."""
        return self.is_not_null().not_()

    def is_not_null(self) -> Timestream:
        """Return a boolean Timestream containing `true` when `self` is not `null`."""
        return Timestream._call("is_valid", self)

    def filter(self, condition: Arg) -> Timestream:
        """Return a Timestream containing only the points where `condition` is `true`.

        Args:
            condition: The condition to filter on.
        """
        return Timestream._call("when", condition, self)

    def collect(
        self,
        *,
        max: Optional[int],
        min: Optional[int] = 0,
        window: Optional[kd.windows.Window] = None,
    ) -> Timestream:
        """Return a Timestream collecting up to the last `max` values in the `window`.

        Collects the values for each key separately.

        Args:
            max: The maximum number of values to collect.
              If `None` all values are collected.
            min: The minimum number of values to collect before producing a value.
              Defaults to 0.
            window: The window to use for the aggregation. If not specified,
              the entire Timestream is used.

        Returns:
            A Timestream containing the list of collected elements at each point.
        """
        if pa.types.is_list(self.data_type):
            return (
                record({"value": self})
                .collect(max=max, min=min, window=window)
                .col("value")
            )
        else:
            return _aggregation("collect", self, window, max, min)

    def time(self) -> Timestream:
        """Return a Timestream containing the time of each point."""
        return Timestream._call("time_of", self)

    def lag(self, n: int) -> Timestream:
        """Return a Timestream containing the value `n` points before each point.

        Args:
            n: The number of points to lag by.
        """
        # hack to support structs/lists (as collect supports lists)
        return self.collect(max=n + 1, min=n + 1)[0]

    def if_(self, condition: Arg) -> Timestream:
        """Return a `Timestream` from `self` at points where `condition` is `true`, and `null` otherwise.

        Args:
            condition: The condition to check.
        """
        return Timestream._call("if", condition, self, input=self)

    def null_if(self, condition: Arg) -> Timestream:
        """Return a `Timestream` from `self` at points where `condition` is not `false`, and `null` otherwise.

        Args:
            condition: The condition to check.
        """
        return Timestream._call("null_if", condition, self, input=self)

    def length(self) -> Timestream:
        """Return a Timestream containing the length of `self`.

        Raises:
            TypeError: When the Timestream is not a string or list.
        """
        if self.data_type.equals(pa.string()):
            return Timestream._call("len", self)
        elif isinstance(self.data_type, pa.ListType):
            return Timestream._call("list_len", self)
        else:
            raise TypeError(f"length not supported for {self.data_type}")

    def with_key(self, key: Arg, grouping: Optional[str] = None) -> Timestream:
        """Return a Timestream with a new grouping by `key`.

        Args:
            key: The new key to use for the grouping.
            grouping: A string literal naming the new grouping. If no `grouping` is specified,
              one will be computed from the type of the `key`.
        """
        return Timestream._call("with_key", key, self, grouping, input=self)

    def lookup(self, key: Arg) -> Timestream:
        """Return a Timestream looking up the value of `self` for each `key`.

        For each non-`null` point in the `key` timestream, returns the value
        from `self` at that time and associated with that `key`. Returns `null`
        if the `key` is `null` or if there is no `value` computed for that key
        at the corresponding time.

        Args:
            key: The foreign key to lookup. This must match the type of the keys in `self`.
        """
        return Timestream._call("lookup", key, self, input=self)

    def coalesce(self, arg: Arg, *args: Arg) -> Timestream:
        """Return a Timestream containing the first non-null point from self and the arguments.

        Args:
            arg: The next value to be coalesced (required).
            args: Additional values to be coalesced (optional).

        Returns:
            Timestream containing the first non-null value from each point.
            If all values are null, then returns null.
        """
        return Timestream._call("coalesce", self, arg, *args, input=self)

    def shift_to(self, time: Arg) -> Timestream:
        """Return a Timestream shifting each point forward to `time`.

        If multiple values are shifted to the same time, they will be emitted in
        the order in which they originally occurred.

        Args:
            time: The time to shift to. This must be a datetime or a Timestream of timestamp_ns.

        Raises:
            NotImplementedError: When `time` is a datetime (shift_to literal not yet implemented).
        """
        if isinstance(time, datetime):
            # session = self._ffi_expr.session()
            # time_ns = time.timestamp() * 1e9
            # time_ns = Timestream._literal(time_ns, session=session)
            # time_ns = Timestream.cast(time_ns, pa.timestamp('ns'))
            # return Timestream._call("shift_to", time_ns, self)
            raise NotImplementedError("shift_to with datetime literal unsupported")
        else:
            return Timestream._call("shift_to", time, self, input=self)

    def shift_by(self, delta: Arg) -> Timestream:
        """Return a Timestream shifting each point forward by the `delta`.

        If multiple values are shifted to the same time, they will be emitted in
        the order in which they originally occurred.

        Args:
            delta: The delta to shift the point forward by.
        """
        if isinstance(delta, timedelta):
            session = self._ffi_expr.session()
            seconds = Timestream._call(
                "seconds", int(delta.total_seconds()), session=session
            )
            return Timestream._call("shift_by", seconds, self, input=self)
        else:
            return Timestream._call("shift_by", delta, self, input=self)

    def shift_until(self, predicate: Arg) -> Timestream:
        """Return a Timestream shifting each point forward to the next time `predicate` is true.

        Note that if the `predicate` evaluates to true at the same time as `self`,
        the point will be emitted at that time.

        If multiple values are shifted to the same time, they will be emitted in
        the order in which they originally occurred.

        Args:
            predicate: The predicate to determine whether to emit shifted rows.
        """
        return Timestream._call("shift_until", predicate, self, input=self)

    def sum(self, *, window: Optional[kd.windows.Window] = None) -> Timestream:
        """Return a Timestream summing the values in the `window`.

        Computes the sum for each key separately.

        Args:
            window: The window to use for the aggregation. Defaults to the entire Timestream.
        """
        return _aggregation("sum", self, window)

    def first(self, *, window: Optional[kd.windows.Window] = None) -> Timestream:
        """Return a Timestream containing the first value in the `window`.

        Computed for each key separately.

        Args:
            window: The window to use for the aggregation. Defaults to the entire Timestream.
        """
        return _aggregation("first", self, window)

    def last(self, window: Optional[kd.windows.Window] = None) -> Timestream:
        """Return a Timestream containing the last value in the `window`.

        Computed for each key separately.

        Args:
            window: The window to use for the aggregation. Defaults to the entire Timestream.
        """
        return _aggregation("last", self, window)

    def count(self, window: Optional[kd.windows.Window] = None) -> Timestream:
        """Return a Timestream containing the count value in the `window`.

        Computed for each key separately.

        Args:
            window: The window to use for the aggregation. Defaults to the entire Timestream.
        """
        return _aggregation("count", self, window)

    def count_if(self, window: Optional[kd.windows.Window] = None) -> Timestream:
        """Return a Timestream containing the count of `true` values in `window`.

        Computed for each key separately.

        Args:
            window: The window to use for the aggregation. Defaults to the entire Timestream.
        """
        return _aggregation("count_if", self, window)

    def max(self, window: Optional[kd.windows.Window] = None) -> Timestream:
        """Return a Timestream containing the max value in the `window`.

        Computed for each key separately.

        Args:
            window: The window to use for the aggregation. Defaults to the entire Timestream.

        See Also:
            This returns the maximum of values in a column. See
            :func:`greatest` to get the maximum value
            between Timestreams at each point.
        """
        return _aggregation("max", self, window)

    def min(self, window: Optional[kd.windows.Window] = None) -> Timestream:
        """Return a Timestream containing the min value in the `window`.

        Computed for each key separately.

        Args:
            window: The window to use for the aggregation. Defaults to the entire Timestream.

        See Also:
            This returns the minimum of values in a column. See
            :func:`least` to get the minimum value
            between Timestreams at each point.
        """
        return _aggregation("min", self, window)

    def mean(self, window: Optional[kd.windows.Window] = None) -> Timestream:
        """Return a Timestream containing the mean value in the `window`.

        Computed for each key separately.

        Args:
            window: The window to use for the aggregation. Defaults to the entire Timestream.
        """
        return _aggregation("mean", self, window)

    def stddev(self, window: Optional[kd.windows.Window] = None) -> Timestream:
        """Return a Timestream containing the standard deviation in the `window`.

        Computed for each key separately.

        Args:
            window: The window to use for the aggregation. Defaults to the entire Timestream.
        """
        return _aggregation("stddev", self, window)

    def variance(self, window: Optional[kd.windows.Window] = None) -> Timestream:
        """Return a Timestream containing the variance in the `window`.

        Computed for each key separately.

        Args:
            window: The window to use for the aggregation. Defaults to the entire Timestream.
        """
        return _aggregation("variance", self, window)

    def cast(self, data_type: pa.DataType) -> Timestream:
        """Return this Timestream cast to the given data type.

        Args:
            data_type: The DataType to cast to.
        """
        return Timestream(self._ffi_expr.cast(data_type))

    def else_(self, other: Timestream) -> Timestream:
        """Return a Timestream containing `self` when not `null`, and `other` otherwise.

        Args:
            other: The Timestream to use if self is `null`.
        """
        return Timestream._call("else", other, self)

    def seconds_since(self, time: Arg) -> Timestream:
        """Return a Timestream containing seconds between `time` and `self`.

        If `self.time()` is greater than `time`, the result will be positive.

        Args:
            time: The time to compute the seconds since.

                This can be either a stream of timestamps or a datetime literal.
                If `time` is a Timestream, the result will contain the seconds
                from `self.time()` to `time.time()` for each point.
        """
        if isinstance(time, datetime):
            session = self._ffi_expr.session()
            nanos = Timestream._literal(time.timestamp() * 1e9, session=session)
            nanos = Timestream.cast(nanos, pa.timestamp("ns", None))
            return Timestream._call("seconds_between", nanos, self)
        else:
            return Timestream._call("seconds_between", time, self)

    def seconds_since_previous(self, n: int = 1) -> Timestream:
        """Return a Timestream containing seconds between `self` and the time `n` points ago.

        Args:
            n: The number of points to look back. For example, `n=1` refers to
              the previous point.

              Defaults to 1 (the previous point).
        """
        time_of_current = Timestream._call("time_of", self).cast(pa.int64())
        time_of_previous = Timestream._call("time_of", self).lag(n).cast(pa.int64())

        # `time_of` returns nanoseconds, so divide to get seconds
        return time_of_current.sub(time_of_previous).div(1e9).cast(pa.duration("s"))

    def flatten(self) -> Timestream:
        """Flatten a list of lists to a list of values."""
        return Timestream._call("flatten", self)

    def union(self, other: Arg) -> Timestream:
        """Union the lists in this timestream with the lists in the other Timestream.

        This corresponds to a pair-wise union within each row of the timestreams.

        Args:
            other: The Timestream of lists to union with.
        """
        return Timestream._call("union", self, other, input=self)

    def record(self, fields: Callable[[Timestream], Mapping[str, Arg]]) -> Timestream:
        """Return a record Timestream from fields computed from this timestream.

        Args:
            fields: The fields to include in the record.

        See Also:
            kaskada.record: Function for creating a record from one or more
              timestreams.
        """
        return record(fields(self))

    def round(self) -> Timestream:
        """Return a Timestream with all values rounded to the nearest integer.

        Returns:
            A Timestream of the same type as `self`. The result contains `null`
            if the value was `null` at that point. Otherwise, it contains
            the result of rounding the value to the nearest integer.

        Notes:
            This method may be applied to any numeric type. For anything other
            than `float32` and `float64` it has no affect since the values
            are already integers.

        See Also:
            - :func:`ceil`
            - :func:`floor`
        """
        return Timestream._call("round", self)

    def sqrt(self) -> Timestream:
        """Return a Timestream with the square root of all values."""
        return Timestream._call("sqrt", self)

    def upper(self) -> Timestream:
        """Return a Timestream with all values converted to upper case."""
        return Timestream._call("upper", self)

    def greatest(self, rhs: Arg) -> Timestream:
        """Return a Timestream with the maximum value of `self` and `rhs` at each point.

        Args:
            rhs: The Timestream or literal value to compare to this.

        Returns:
            Each point contains the value from `self` if `self`
            is greater than `rhs`, otherwise it contains `rhs`.
            If any input is `null` or `NaN`, then that will be
            the result.

        See Also:
            This returns the greatest of two values. See
            :func:`max` for the maximum of values in
            a column.
        """
        return Timestream._call("zip_max", self, rhs)

    def least(self, rhs: Arg) -> Timestream:
        """Return a Timestream with the minimum value of `self` and `rhs` at each point.

        Args:
            rhs: The Timestream or literal value to compare to this.

        Returns:
            Each point contains the value from `self` if `self`
            is less than `rhs`, otherwise it contains `rhs`.
            If any input is `null` or `NaN`, then that will be
            the result.

        See Also:
            This returns the least of two values. See
            :func:`min` for the minimum of values in
            a column.
        """
        return Timestream._call("zip_min", self, rhs)

    def preview(
        self,
        limit: int = 10,
        results: Optional[Union[kd.results.History, kd.results.Snapshot]] = None,
    ) -> pd.DataFrame:
        """Preview the points in this TimeStream as a DataFrame.

        Args:
            limit: The number of points to preview.
            results: The results to produce. Defaults to `History()` producing all points.
        """
        return self.to_pandas(results, row_limit=limit)

    def to_pandas(
        self,
        results: Optional[Union[kd.results.History, kd.results.Snapshot]] = None,
        *,
        row_limit: Optional[int] = None,
    ) -> pd.DataFrame:
        """Execute the TimeStream with the given options and return a DataFrame.

        Args:
            results: The results to produce in the DataFrame. Defaults to `History()` producing all points.
            row_limit: The maximum number of rows to return. Defaults to `None` for no limit.
            max_batch_size: The maximum number of rows to return in each batch.
              Defaults to `None` for no limit.

        See Also:
            - :func:`preview`: For quick peeks at the contents of a TimeStream during development.
            - :func:`write`: For writing results to supported destinations without passing through
              Pandas.
            - :func:`run_iter`: For non-blocking (iterator or async iterator) execution.
        """
        execution = self._execute(results, row_limit=row_limit)
        batches = execution.collect_pyarrow()
        table = pa.Table.from_batches(batches, schema=execution.schema())
        table = table.drop_columns(["_subsort", "_key_hash"])
        return table.to_pandas()

    def write(
        self,
        destination: kd.destinations.Destination,
        mode: Literal["once", "live"] = "once",
        results: Optional[Union[kd.results.History, kd.results.Snapshot]] = None,
    ) -> Execution:
        """Execute the TimeStream writing to the given destination.

        Args:
            destination: The destination to write to.
            mode: The execution mode to use. Defaults to `'once'` to produce the results
              from the currently available data. Use `'live'` to start a standing query
              that continues to process new data until stopped.
            results: The results to produce. Defaults to `History()` producing all points.

        Returns:
            An `ExecutionProgress` which allows iterating (synchronously or asynchronously)
            over the progress information, as well as cancelling the query if it is no longer
            needed.
        """
        raise NotImplementedError

    @overload
    def run_iter(
        self,
        kind: Literal["pandas"] = "pandas",
        *,
        mode: Literal["once", "live"] = "once",
        results: Optional[Union[kd.results.History, kd.results.Snapshot]] = None,
        row_limit: Optional[int] = None,
        max_batch_size: Optional[int] = None,
    ) -> ResultIterator[pd.DataFrame]:
        ...

    @overload
    def run_iter(
        self,
        kind: Literal["pyarrow"],
        *,
        mode: Literal["once", "live"] = "once",
        results: Optional[Union[kd.results.History, kd.results.Snapshot]] = None,
        row_limit: Optional[int] = None,
        max_batch_size: Optional[int] = None,
    ) -> ResultIterator[pa.RecordBatch]:
        ...

    @overload
    def run_iter(
        self,
        kind: Literal["row"],
        *,
        mode: Literal["once", "live"] = "once",
        results: Optional[Union[kd.results.History, kd.results.Snapshot]] = None,
        row_limit: Optional[int] = None,
        max_batch_size: Optional[int] = None,
    ) -> ResultIterator[dict]:
        ...

    def run_iter(
        self,
        kind: Literal["pandas", "pyarrow", "row"] = "pandas",
        *,
        mode: Literal["once", "live"] = "once",
        results: Optional[Union[kd.results.History, kd.results.Snapshot]] = None,
        row_limit: Optional[int] = None,
        max_batch_size: Optional[int] = None,
    ) -> Union[
        ResultIterator[pd.DataFrame],
        ResultIterator[pa.RecordBatch],
        ResultIterator[dict],
    ]:
        """Execute the TimeStream producing an iterator over the results.

        Args:
            kind: The kind of iterator to produce. Defaults to `pandas`.
            mode: The execution mode to use. Defaults to `'once'` to produce the results
              from the currently available data. Use `'live'` to start a standing query
              that continues to process new data until stopped.
            results: The results to produce. Defaults to `History()` producing all points.
            row_limit: The maximum number of rows to return. Defaults to `None` for no limit.
            max_batch_size: The maximum number of rows to return in each batch.
              Defaults to `None` for no limit.

        Returns:
            Iterator over data of the corresponding kind. The `QueryIterator` allows
            cancelling the query or materialization as well as iterating.

        See Also:
            - :func:`write`: To write the results directly to a
              :class:`Destination<kaskada.destinations.Destination>`.
        """
        execution = self._execute(
            results, mode=mode, row_limit=row_limit, max_batch_size=max_batch_size
        )
        if kind == "pandas":
            return ResultIterator(execution, lambda table: iter((table.to_pandas(),)))
        elif kind == "pyarrow":
            return ResultIterator(execution, lambda table: table.to_batches())
        elif kind == "row":
            return ResultIterator(execution, lambda table: iter(table.to_pylist()))

        raise AssertionError(f"Unhandled kind {kind}")

    def explain(
        self,
        kind: Literal["initial_dfg", "final_dfg", "final_plan"] = "final_plan",
        results: Optional[Union[kd.results.History, kd.results.Snapshot]] = None,
        mode: Literal["once", "live"] = "once",
    ) -> "graphviz.Source":
        """Return an explanation of this Timestream will be executed.

        This is intended for understanding how a given Timestream query will
        be executed. It does not guarantee that the same query will always
        be executed the same way.

        Args:
            results: The results to produce. Defaults to `History()` producing all points.
            kind: The kind of plan to produce.
            mode: The execution mode to use. Defaults to `'once'` to produce the results
              from the currently available data. Use `'live'` to start a standing query
              that continues to process new data until stopped.

        Returns:
            A GraphViz representation of the execution plan as a string, SVG string, or SVG.
            Specific representation depends on the `format` argument.

        Raises:
            ValueError if the `kind` is not recognized or the `format` is not supported.

        Caution:
            This method is intended for debugging and development purposes only.
            The API may change in the future.
        """
        import graphviz

        expr = self
        if not pa.types.is_struct(self.data_type):
            # The execution engine requires a struct, so wrap this in a record.
            expr = record({"result": self})

        options = _ExecutionOptions.create(
            results, row_limit=None, max_batch_size=None, mode=mode
        )

        dot = expr._ffi_expr.explain(kind, options)
        dot = graphviz.Source(dot, engine="dot")
        return dot

    def _execute(
        self,
        results: Optional[Union[kd.results.History, kd.results.Snapshot]],
        *,
        row_limit: Optional[int] = None,
        max_batch_size: Optional[int] = None,
        mode: Literal["once", "live"] = "once",
    ) -> _ffi.Execution:
        """Execute the timestream and return the FFI handle for the execution."""
        expr = self
        if not pa.types.is_struct(self.data_type):
            # The execution engine requires a struct, so wrap this in a record.
            expr = record({"result": self})

        options = _ExecutionOptions.create(results, row_limit, max_batch_size, mode)
        return expr._ffi_expr.execute(options)


def _aggregation(
    op: str,
    input: Timestream,
    window: Optional[kd.windows.Window],
    *args: Union[Timestream, LiteralValue],
) -> Timestream:
    """Return the aggregation `op` with the given `input`, `window` and `args`.

    Args:
    op: The operation to create.
    input: The input to the aggregation.
    window: The window to use for the aggregation.
    *args: Additional arguments to provide after `input` and before the flattened window.

    Raises:
        NotImplementedError: If the window is not a known type.
    """
    if window is None:
        return Timestream._call(op, input, *args, None, None)
    elif isinstance(window, kd.windows.Since):
        predicate = window.predicate
        if callable(predicate):
            predicate = predicate(input)
        return Timestream._call(op, input, *args, predicate, None)
    elif isinstance(window, kd.windows.Sliding):
        predicate = window.predicate
        if callable(predicate):
            predicate = predicate(input)
        # Sliding windows produce non-cumulative values, hence the filter at the end.
        return Timestream._call(op, input, *args, predicate, window.duration).filter(
            predicate
        )
    elif isinstance(window, kd.windows.Trailing):
        if op != "collect":
            raise NotImplementedError(
                f"Aggregation '{op} does not support trailing windows"
            )

        trailing_ns = int(window.duration.total_seconds() * 1e9)

        # Create the shifted-forward input
        input_shift = input.shift_by(window.duration)

        # Merge, then extract just the input.
        #
        # Note: Assumes the "input" is discrete. Can probably
        # use a transform to make it discrete (eg., `input.filter(True)`)
        # or a special function to do that.
        #
        # HACK: This places an extra null row in the input to `collect`
        # which allows us to "clear" the window when the appropriate
        # `duration` has passed with no "real" inputs.
        merged_input = record({"input": input, "shift": input_shift}).col("input")
        return Timestream._call("collect", merged_input, *args, None, trailing_ns)
    elif isinstance(window, kd.windows.Tumbling):
        # Tumbling windows are analogous to Since windows, aside from output behavior.
        # Tumbling windows only emit once per window. However, this behavior is not implemented
        # in Sparrow yet, so we hack this by using a Since window with a filter applied afterwards
        # with the same predicate. Note this hack is brittle, adds additional work with merging and filters,
        # and generally is not how we want to handle new window behaviors.
        predicate = window.predicate
        if callable(predicate):
            predicate = predicate(input)

        return Timestream._call(op, input, *args, predicate, None).filter(predicate)
    else:
        raise NotImplementedError(f"Unknown window type {window!r}")


def record(fields: Mapping[str, Arg]) -> Timestream:
    """Return a record Timestream from the given fields.

    Args:
        fields: The fields to include in the record.

    See Also:
        Timestream.record: Method for creating a record from fields computed from
          a timestream.
    """
    import itertools

    args: List[Arg] = list(itertools.chain(*fields.items()))
    return Timestream._call("record", *args)



import kaskada as kd


async def test_read_jsonl(golden) -> None:
    source = await kd.sources.JsonlString.create(
        '{"time": "1996-12-19T16:39:57", "user": "A", "m": 5, "n": 10}',
        time_column="time",
        key_column="user",
    )
    golden.jsonl(source)

    await source.add_string(
        """
        {"time": "1996-12-19T16:40:57", "user": "A", "m": 8, "n": 10}
        {"time": "1996-12-19T16:41:57", "user": "B", "m": 5}
        """
    )

    golden.jsonl(source)


async def test_read_jsonl_lists(golden) -> None:
    source = await kd.sources.JsonlString.create(
        '{"time": "1996-12-19T16:39:57", "user": "A", "m": [5, 10], "n": 10}',
        time_column="time",
        key_column="user",
    )
    golden.jsonl(source)

    await source.add_string(
        """
        {"time": "1996-12-19T16:40:57", "user": "A", "m": [], "n": 10}
        {"time": "1996-12-19T16:41:57", "user": "A", "n": 10}
        """
    )
    golden.jsonl(source)


async def test_read_jsonl_with_subsort(golden) -> None:
    source = await kd.sources.JsonlString.create(
        '{"time": "1996-12-19T16:39:57", "user": "A", "m": 5, "n": 10, "subsort": 1}',
        time_column="time",
        key_column="user",
        subsort_column="subsort",
    )
    golden.jsonl(source)

    await source.add_string(
        """
        {"time": "1996-12-19T16:40:57", "user": "A", "m": 8, "n": 10, "subsort": 2}
        {"time": "1996-12-19T16:41:57", "user": "B", "m": 5, "subsort": 3}
        """
    )

    golden.jsonl(source)



import asyncio
import os
from typing import Union

import kaskada as kd
import pandas as pd
import pytest
from kaskada import init_session


pytest_plugins = ["pytest_docker_fixtures"]


def pytest_addoption(parser: pytest.Parser):
    parser.addoption("--save-golden", action="store_true", help="update golden files")
    parser.addoption(
        "-S",
        action="append",
        metavar="SVC",
        help="run tests requiring the service SVC.",
    )


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers", "svc(name): tests that require the named service"
    )


def pytest_runtest_setup(item):
    """Skip the test unless all of the marked services are present."""
    required_svcs = {mark.args[0] for mark in item.iter_markers(name="svc")}
    provided_svcs = set(item.config.getoption("-S") or [])

    missing_svcs = required_svcs - provided_svcs
    if missing_svcs:
        pytest.skip(f"test requires services {missing_svcs!r}")


@pytest.fixture(autouse=True, scope="session")
def session() -> None:
    init_session()


# sets up a single, session-scoped async event loop.
@pytest.fixture(scope="session")
def event_loop():
    policy = asyncio.get_event_loop_policy()
    loop = policy.new_event_loop()
    yield loop
    loop.close()


class GoldenFixture(object):
    def __init__(self, dirname: str, test_name: str, save: bool):
        self._output = 0
        self._dirname = dirname
        self._test_name = test_name
        self._save = save

    def jsonl(self, data: Union[kd.Timestream, pd.DataFrame]) -> None:
        """Golden test against newline-delimited JSON file (json-lines)."""
        df = _data_to_dataframe(data)
        filename = self._filename("jsonl")

        if self._save:
            df.to_json(
                filename,
                orient="records",
                lines=True,
                date_format="iso",
                date_unit="ns",
            )

        golden = pd.read_json(
            filename,
            orient="records",
            lines=True,
            dtype=df.dtypes.to_dict(),
            date_unit="ns",
        )

        if golden.empty and df.empty:
            # For some reason, even when specifying the dtypes, reading an empty
            # json produces no columns.
            return
        pd.testing.assert_frame_equal(df, golden, check_datetimelike_compat=True)

    def parquet(self, data: Union[kd.Timestream, pd.DataFrame]) -> None:
        """Golden test against Parquet file."""
        df = _data_to_dataframe(data)
        filename = self._filename("parquet")

        if self._save:
            df.to_parquet(filename)

        golden = pd.read_parquet(filename)

        pd.testing.assert_frame_equal(df, golden)

    def _filename(self, suffix: str) -> str:
        filename = (
            f"{self._test_name}.{suffix}"
            if self._output == 0
            else f"{self._test_name}_{self._output}.{suffix}"
        )
        filename = os.path.join(self._dirname, filename)
        self._output += 1

        if not self._save:
            assert os.path.exists(
                filename
            ), f"Golden file {filename} does not exist. Run with `--save-golden` to create it."
        return filename


def _data_to_dataframe(data: Union[kd.Timestream, pd.DataFrame]) -> pd.DataFrame:
    if isinstance(data, pd.DataFrame):
        return data
    elif isinstance(data, kd.Timestream):
        return data.to_pandas()
    else:
        raise ValueError(f"data must be a Timestream or a DataFrame, was {type(data)}")


@pytest.fixture
def golden(
    request: pytest.FixtureRequest, pytestconfig: pytest.Config
) -> GoldenFixture:
    """Test fixture for checking results against a golden file."""
    test_name = request.node.name
    module_name = request.node.module.__name__
    dirname = os.path.join("pytests", "golden", module_name)

    save = pytestconfig.getoption("--save-golden", default=False)
    if save:
        os.makedirs(dirname, exist_ok=True)
    else:
        assert os.path.isdir(
            dirname
        ), f"golden directory {dirname} does not exist. run with `--save-golden` to create it."

    return GoldenFixture(dirname, test_name, save)



"""Sources of data for Kaskada queries."""
from .arrow import CsvString, JsonlFile, JsonlString, Pandas, Parquet, PyDict
from .source import Source


__all__ = [
    "Source",
    "CsvString",
    "Pandas",
    "JsonlFile",
    "JsonlString",
    "PyDict",
    "Parquet",
]



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def key_source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,state",
            "1996-12-19T16:39:57,A,WA",
            "1996-12-19T16:39:58,B,NC",
            "1996-12-19T16:39:59,A,WA",
            "1996-12-19T16:40:00,A,NC",
            "1996-12-19T16:40:01,A,SC",
            "1996-12-19T16:40:02,A,WA",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


@pytest.fixture(scope="module")
async def foreign_source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,NC,10,5",
            "1996-12-19T16:39:58,NC,24,3",
            "1996-12-19T16:39:59,WA,17,6",
            "1996-12-19T16:40:00,NC,,9",
            "1996-12-19T16:40:01,WA,12,",
            "1996-12-19T16:40:02,WA,,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_lookup(key_source, foreign_source, golden) -> None:
    state = key_source.col("state")
    foreign_value = foreign_source.col("m")
    last_foreign_value = foreign_source.col("m").last()
    golden.jsonl(
        kd.record(
            {
                "state": state,
                "lookup": foreign_value.lookup(state),
                "lookup_last": last_foreign_value.lookup(state),
            }
        )
    )



import sys

import kaskada as kd
import pyarrow as pa
import pytest


@pytest.fixture(scope="module")
def source1() -> kd.sources.Source:
    schema = pa.schema(
        [
            pa.field("time", pa.int32(), nullable=False),
            pa.field("key", pa.int64(), nullable=False),
            pa.field("x", pa.float64()),
            pa.field("y", pa.int32()),
        ]
    )
    return kd.sources.Source(schema, time_column="time", key_column="key")


def test_field_ref(source1) -> None:
    field_ref_long = source1.col("x")
    assert field_ref_long.data_type == pa.float64()


def test_field_ref_no_such_field(source1) -> None:
    with pytest.raises(ValueError, match="Illegal field reference"):
        # This raises a "NoSuchAttribute" error.
        # We currently catch this in Python and don't do anything to
        # suggest possible alternatives.
        #
        # TODO: We should either surface the Sparrow error which suggests
        # possible field names, or improve the Python error.
        source1.col("foo")


def test_field_ref_not_a_struct(source1) -> None:
    with pytest.raises(
        TypeError, match="Cannot access column 'x' of non-record type 'double'"
    ):
        source1.col("x").col("x")


def test_timestream_math(source1) -> None:
    x = source1.col("x")
    assert (x + 1).data_type == x.data_type
    assert (1 + x).data_type == x.data_type
    assert (x - 1).data_type == x.data_type
    assert (1 - x).data_type == x.data_type
    assert (x * 1).data_type == x.data_type
    assert (1 * x).data_type == x.data_type
    assert (1 / x).data_type == x.data_type
    assert (1 + x).data_type == x.data_type

    assert (1 + x).neg().data_type == x.data_type


def test_timestream_comparison(source1) -> None:
    x = source1.col("x")

    # Tests the various comparison operators. Even though Python doesn't have a
    # `__rgt__` (reverse gt) dunder method, if the LHS doesn't support `gt` with
    # the RHS it seems to try `rhs lt lhs`.
    assert (x > 1).data_type == pa.bool_()
    assert (1 > x).data_type == pa.bool_()
    assert (x < 1).data_type == pa.bool_()
    assert (1 < x).data_type == pa.bool_()
    assert (x >= 1).data_type == pa.bool_()
    assert (1 >= x).data_type == pa.bool_()
    assert (x <= 1).data_type == pa.bool_()
    assert (1 <= x).data_type == pa.bool_()

    # For `eq` and `ne` we only support timestream on the LHS since it is a method.
    # We can't overload `__eq__` since that must take any RHS and must return `bool`.
    assert x.eq(1).data_type == pa.bool_()
    assert x.ne(1).data_type == pa.bool_()

    a = x > 1
    b = x.eq(1)
    assert a.not_().data_type == pa.bool_()
    assert a.and_(b).data_type == pa.bool_()
    assert a.or_(b).data_type == pa.bool_()


def test_timestream_eq_warning(source1) -> None:
    x = source1.col("x")
    y = source1.col("y")

    with pytest.warns(UserWarning, match="Use 'eq'"):
        x == y  # noqa : B015

    with pytest.warns(UserWarning, match="Use 'ne'"):
        x != y  # noqa : B015


def test_timestream_arithmetic_types(source1) -> None:
    x = source1.col("x")
    assert (x.eq(1)).data_type == pa.bool_()
    assert (x + 1).data_type == pa.float64()
    assert (x + source1.col("y")).data_type == pa.float64()

    # TODO: This should raise a TypeError, but currently the Rust
    # code always raises a ValueError, so everything comes out
    # looking the same.
    with pytest.raises(ValueError) as e:
        x.eq(1) + source1.col("y")
    assert "Incompatible argument types" in str(e)
    if sys.version_info >= (3, 11):
        assert "Arg[0]: Timestream[bool]" in e.value.__notes__
        assert "Arg[1]: Timestream[int32]" in e.value.__notes__


async def test_timestream_preview(golden) -> None:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1996-12-19T16:40:00,A,,9",
            "1996-12-19T16:40:01,A,12,",
            "1996-12-19T16:40:02,A,,",
        ]
    )
    source = await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )

    golden.jsonl(source.preview(limit=4))


async def test_timestream_run_non_record(golden) -> None:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1996-12-19T16:40:00,A,,9",
            "1996-12-19T16:40:01,A,12,",
            "1996-12-19T16:40:02,A,,",
        ]
    )
    source = await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )
    golden.jsonl(source.col("m"))


async def test_timestream_cast(golden) -> None:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1996-12-19T16:40:00,A,,9",
            "1996-12-19T16:40:01,A,12,",
            "1996-12-19T16:40:02,A,,",
        ]
    )
    source = await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )
    golden.jsonl(source.col("time").cast(pa.timestamp("ns")))


def test_timestream_call_all_literals() -> None:
    with pytest.raises(ValueError, match="at least one non-literal"):
        kd.Timestream._call("foo", 1, 2, 3)



from datetime import timedelta

import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n,s,b",
            "1996-12-19T16:39:57,A,5,10,a,true",
            "1996-12-19T16:39:58,B,24,3,b,true",
            "1996-12-19T16:39:59,A,17,6,b,",
            "1996-12-19T16:40:00,A,,9,,false",
            "1996-12-19T16:40:01,A,12,,e,false",
            "1996-12-19T16:40:02,A,,,f,true",
            "1996-12-19T16:40:04,A,,,f,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_collect_basic(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "collect_m": m.collect(max=None),
                "n": n,
                "collect_n": n.collect(max=None),
            }
        )
    )


async def test_collect_with_max(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "collect_m_max_2": m.collect(max=2),
                "n": n,
                "collect_n_max_2": n.collect(max=2),
            }
        )
    )


async def test_collect_with_min(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "collect_m_min_2": m.collect(min=2, max=None),
                "n": n,
                "collect_n_min_2": n.collect(min=2, max=None),
            }
        )
    )


async def test_collect_with_min_and_max(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "collect_m_min_2_max_2": m.collect(min=2, max=2),
                "n": n,
                "collect_n_min_2_max_2": n.collect(min=2, max=2),
            }
        )
    )


async def test_collect_since_window(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(
        kd.record(
            {"m": m, "since_m": m.collect(max=None, window=kd.windows.Since(m > 10))}
        )
    )


async def test_collect_i64_trailing_window_1s(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "collect_m": m.collect(
                    max=None, window=kd.windows.Trailing(timedelta(seconds=1))
                ),
            }
        )
    )


async def test_collect_i64_trailing_window_3s(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "collect_m": m.collect(
                    max=None, window=kd.windows.Trailing(timedelta(seconds=3))
                ),
            }
        )
    )


async def test_collect_i64_trailing_window_3s_with_max(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "collect_m": m.collect(
                    max=2, window=kd.windows.Trailing(timedelta(seconds=3))
                ),
            }
        )
    )


async def test_collect_i64_trailing_window_3s_with_min(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "collect_m": m.collect(
                    min=3, max=None, window=kd.windows.Trailing(timedelta(seconds=3))
                ),
            }
        )
    )


async def test_collect_string_trailing_window_1s(source, golden) -> None:
    s = source.col("s")
    golden.jsonl(
        kd.record(
            {
                "m": s,
                "collect_s": s.collect(
                    max=None, window=kd.windows.Trailing(timedelta(seconds=1))
                ),
            }
        )
    )


async def test_collect_string_trailing_window_3s(source, golden) -> None:
    s = source.col("s")
    golden.jsonl(
        kd.record(
            {
                "s": s,
                "collect_s": s.collect(
                    max=None, window=kd.windows.Trailing(timedelta(seconds=3))
                ),
            }
        )
    )


async def test_collect_string_trailing_window_3s_with_max(source, golden) -> None:
    s = source.col("s")
    golden.jsonl(
        kd.record(
            {
                "s": s,
                "collect_s": s.collect(
                    max=2, window=kd.windows.Trailing(timedelta(seconds=3))
                ),
            }
        )
    )


async def test_collect_string_trailing_window_3s_with_min(source, golden) -> None:
    s = source.col("s")
    golden.jsonl(
        kd.record(
            {
                "s": s,
                "collect_s": s.collect(
                    min=3, max=None, window=kd.windows.Trailing(timedelta(seconds=3))
                ),
            }
        )
    )


async def test_collect_bool_trailing_window_1s(source, golden) -> None:
    b = source.col("b")
    golden.jsonl(
        kd.record(
            {
                "b": b,
                "collect_b": b.collect(
                    max=None, window=kd.windows.Trailing(timedelta(seconds=1))
                ),
            }
        )
    )


async def test_collect_bool_trailing_window_3s(source, golden) -> None:
    b = source.col("b")
    golden.jsonl(
        kd.record(
            {
                "b": b,
                "collect_b": b.collect(
                    max=None, window=kd.windows.Trailing(timedelta(seconds=3))
                ),
            }
        )
    )


async def test_collect_bool_trailing_window_3s_with_max(source, golden) -> None:
    b = source.col("b")
    golden.jsonl(
        kd.record(
            {
                "b": b,
                "collect_b": b.collect(
                    max=2, window=kd.windows.Trailing(timedelta(seconds=3))
                ),
            }
        )
    )


async def test_collect_bool_trailing_window_3s_with_min(source, golden) -> None:
    b = source.col("b")
    golden.jsonl(
        kd.record(
            {
                "b": b,
                "collect_b": b.collect(
                    min=3, max=None, window=kd.windows.Trailing(timedelta(seconds=3))
                ),
            }
        )
    )


# Currently, the Pandas comparison method being used doesn't handle
# date-time like fields nested within a list. So we expand things out.
#
# TODO: Improve the golden testing so this isn't necessary.
async def test_collect_struct_trailing_window_1s(source, golden) -> None:
    collect = source.collect(max=None, window=kd.windows.Trailing(timedelta(seconds=1)))
    golden.jsonl(
        kd.record(
            {
                "f0": collect[0].col("time"),
                "f1": collect[1].col("time"),
                "f2": collect[2].col("time"),
                "f3": collect[3].col("time"),
                "f4": collect[4].col("time"),
            }
        )
    )


async def test_collect_struct_trailing_window_3s(source, golden) -> None:
    collect = source.collect(max=None, window=kd.windows.Trailing(timedelta(seconds=3)))
    golden.jsonl(
        kd.record(
            {
                "f0": collect[0].col("time"),
                "f1": collect[1].col("time"),
                "f2": collect[2].col("time"),
                "f3": collect[3].col("time"),
                "f4": collect[4].col("time"),
            }
        )
    )


async def test_collect_struct_trailing_window_3s_with_max(source, golden) -> None:
    collect = source.collect(max=2, window=kd.windows.Trailing(timedelta(seconds=3)))
    golden.jsonl(
        kd.record(
            {
                "f0": collect[0].col("time"),
                "f1": collect[1].col("time"),
                "f2": collect[2].col("time"),
                "f3": collect[3].col("time"),
                "f4": collect[4].col("time"),
            }
        )
    )


async def test_collect_struct_trailing_window_3s_with_min(source, golden) -> None:
    collect = source.collect(
        min=3, max=None, window=kd.windows.Trailing(timedelta(seconds=3))
    )
    golden.jsonl(
        kd.record(
            {
                "f0": collect[0].col("time"),
                "f1": collect[1].col("time"),
                "f2": collect[2].col("time"),
                "f3": collect[3].col("time"),
                "f4": collect[4].col("time"),
            }
        )
    )


async def test_collect_records(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(kd.record({"m": m, "n": n}).collect(max=None))


async def test_collect_records_field_ref(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(kd.record({"m": m, "n": n}).collect(max=None).col("m"))


async def test_collect_lists(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "list_m": m.collect(max=10),
                "collect_list": m.collect(max=10).collect(max=10),
            }
        )
    )



#!/usr/bin/env python
#
# Bluesky Kaskada Consumer
#
# This script demonstrates the use of Kaskada to consume and compute over
# the BlueSky firehose.

import asyncio
import time

import kaskada as kd
import pyarrow as pa
from atproto import CAR, AtUri, models
from atproto.firehose import (
    AsyncFirehoseSubscribeReposClient,
    parse_subscribe_repos_message,
)
from atproto.xrpc_client.models import get_or_create, ids, is_record_type


async def main():
    # Initialize the Kaskada session so we can use it for stream processing
    kd.init_session()

    # Create the BlueSky (At protocol) client.
    # The firehose doesn't (currently) require authentication.
    at_client = AsyncFirehoseSubscribeReposClient()

    # [start_setup]
    # Setup the data source.
    # This defintes (most of) the schema of the events we'll receive,
    # and tells Kaskada which fields to use for time and initial entity.
    #
    # We'll push events into this source as they arrive in real-time.
    posts = kd.sources.PyDict(
        rows=[],
        schema=pa.schema(
            [
                pa.field(
                    "record",
                    pa.struct(
                        [
                            pa.field("created_at", pa.string()),
                            pa.field("text", pa.string()),
                            # pa.field("embed", pa.struct([...])),
                            pa.field("entities", pa.string()),
                            # pa.field("facets", pa.list_(...)),
                            pa.field("labels", pa.float64()),
                            pa.field("langs", pa.list_(pa.string())),
                            # pa.field("reply", pa.struct([...])),
                            pa.field("py_type", pa.string()),
                        ]
                    ),
                ),
                pa.field("uri", pa.string()),
                pa.field("cid", pa.string()),
                pa.field("author", pa.string()),
                pa.field("ts", pa.float64()),
            ]
        ),
        time_column="ts",
        key_column="author",
        time_unit="s",
    )
    # [end_setup]

    # [start_incoming]
    # Handler for newly-arrived messages from BlueSky.
    async def receive_at(message) -> None:
        # Extract the contents of the message and bail if it's not a "commit"
        commit = parse_subscribe_repos_message(message)
        if not isinstance(commit, models.ComAtprotoSyncSubscribeRepos.Commit):
            return

        # Get the operations included in the message
        ops = _get_ops_by_type(commit)
        for new_post in ops["posts"]["created"]:
            # The atproto library's use of schemas is sort of confusing
            # This isn't always the expected type and I'm not sure why...
            if not isinstance(new_post["record"], models.app.bsky.feed.post.Main):
                continue

            # The parsing produces a hot mess of incompatible types, so we build
            # a dict from scratch to simplify.
            posts.add_rows(
                {
                    "record": dict(new_post["record"]),
                    "uri": new_post["uri"],
                    "cid": new_post["cid"],
                    "author": new_post["author"],
                    "ts": time.time(),
                }
            )

    # [end_incoming]

    # [start_result]
    # Handler for values emitted by Kaskada.
    async def receive_outputs():
        # We'll perform a very simple aggregation - key by language and count.
        posts_by_first_lang = posts.with_key(posts.col("record").col("langs").index(0))

        # Consume outputs as they're generated and print to STDOUT.
        async for row in posts_by_first_lang.count().run_iter(kind="row", mode="live"):
            print(f"{row['_key']} has posted {row['result']} times since startup")

    # [end_result]

    # [start_run]
    # Kickoff the two async processes concurrently.
    await asyncio.gather(at_client.start(receive_at), receive_outputs())
    # [end_run]


# Copied from https://raw.githubusercontent.com/MarshalX/atproto/main/examples/firehose/process_commits.py
def _get_ops_by_type(  # noqa: C901, E302
    commit: models.ComAtprotoSyncSubscribeRepos.Commit,
) -> dict:  # noqa: C901, E302
    operation_by_type = {
        "posts": {"created": [], "deleted": []},
        "reposts": {"created": [], "deleted": []},
        "likes": {"created": [], "deleted": []},
        "follows": {"created": [], "deleted": []},
    }

    car = CAR.from_bytes(commit.blocks)
    for op in commit.ops:
        uri = AtUri.from_str(f"at://{commit.repo}/{op.path}")

        if op.action == "update":
            # not supported yet
            continue

        if op.action == "create":
            if not op.cid:
                continue

            create_info = {"uri": str(uri), "cid": str(op.cid), "author": commit.repo}

            record_raw_data = car.blocks.get(op.cid)
            if not record_raw_data:
                continue

            record = get_or_create(record_raw_data, strict=False)
            if uri.collection == ids.AppBskyFeedLike and is_record_type(
                record, ids.AppBskyFeedLike
            ):
                operation_by_type["likes"]["created"].append(
                    {"record": record, **create_info}
                )
            elif uri.collection == ids.AppBskyFeedPost and is_record_type(
                record, ids.AppBskyFeedPost
            ):
                operation_by_type["posts"]["created"].append(
                    {"record": record, **create_info}
                )
            elif uri.collection == ids.AppBskyFeedRepost and is_record_type(
                record, ids.AppBskyFeedRepost
            ):
                operation_by_type["reposts"]["created"].append(
                    {"record": record, **create_info}
                )
            elif uri.collection == ids.AppBskyGraphFollow and is_record_type(
                record, ids.AppBskyGraphFollow
            ):
                operation_by_type["follows"]["created"].append(
                    {"record": record, **create_info}
                )

        if op.action == "delete":
            if uri.collection == ids.AppBskyFeedLike:
                operation_by_type["likes"]["deleted"].append({"uri": str(uri)})
            elif uri.collection == ids.AppBskyFeedPost:
                operation_by_type["posts"]["deleted"].append({"uri": str(uri)})
            elif uri.collection == ids.AppBskyFeedRepost:
                operation_by_type["reposts"]["deleted"].append({"uri": str(uri)})
            elif uri.collection == ids.AppBskyGraphFollow:
                operation_by_type["follows"]["deleted"].append({"uri": str(uri)})

    return operation_by_type


if __name__ == "__main__":
    asyncio.run(main())



import kaskada as kd
import pytest


@pytest.fixture
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1996-12-19T16:40:00,A,,9",
            "1996-12-19T16:40:01,A,12,",
            "1996-12-19T16:40:02,A,,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_record(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")

    golden.jsonl(
        kd.record(
            {
                "m": m,
                "n": n,
            }
        )
    )


async def test_extend_dict(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(source.extend({"add": m + n}))


async def test_extend_record(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(source.extend(kd.record({"add": m + n})))


async def test_extend_computed_record(source, golden) -> None:
    golden.jsonl(source.extend(lambda i: kd.record({"add": i.col("m") + i.col("n")})))


async def test_extend_computed_dict(source, golden) -> None:
    golden.jsonl(source.extend(lambda input: {"add": input.col("m") + input.col("n")}))


async def test_select_record(source, golden) -> None:
    golden.jsonl(source.select("n"))


async def test_remove_record(source, golden) -> None:
    golden.jsonl(source.remove("n"))



"""Render timestreams using Plotly."""

try:
    import plotly.graph_objects as go
    import plotly.io as pio
    from plotly.subplots import make_subplots

    def _require_plotly() -> None:
        pass

except ImportError:

    def _require_plotly() -> None:
        raise ImportError(
            "Plotly is not installed. Install it with `pip install plotly`."
        )


import itertools
from dataclasses import dataclass
from typing import Optional

from ._timestream import Timestream


@dataclass
class Plot(object):
    """Configuration for a single plot to render.

    Attributes:
        stream: The Timestream to render.
        name: The name of the plot to render. Defaults to `Result` if not set.
    """

    stream: Timestream
    name: Optional[str] = None


def render(
    *args: Plot, theme: Optional[str] = None, title_text: Optional[str] = None
) -> None:
    """Render one or more plots."""
    _require_plotly()

    if theme is None:
        theme = "plotly"

    template = pio.templates[theme]

    fig = make_subplots(rows=len(args), cols=1)

    next_color = itertools.cycle(template.layout.colorway)
    color_map = {}

    for row, plot in enumerate(args, 1):
        data = plot.stream.to_pandas()
        name = plot.name or f"Result {row}"

        for key in data["_key"].unique():
            # TODO: Change markers when colorway cycles.
            # TODO: Use different markers for different entity types.
            # TODO: Render categorical values.
            # TODO: Warn if number of points / keys exceeds some threshold.
            new = False
            if key not in color_map:
                new = True
                color_map[key] = next(next_color)
            color = color_map[key]

            points = data[data["_key"] == key]

            # TODO: Render arrows for continuous timestreams.
            # TODO: Render discontinuity at window boundaries.
            fig.append_trace(
                go.Scatter(
                    x=points["_time"],
                    y=points["result"],
                    mode="markers+lines" if plot.stream.is_continuous else "markers",
                    marker=dict(
                        color=color,
                    ),
                    line={"shape": "hv"},
                    legendgroup=key,
                    showlegend=new,
                    name=str(key),
                ),
                row=row,
                col=1,
            )
        fig["layout"][f"xaxis{row}"]["title"] = "Time"
        fig["layout"][f"yaxis{row}"]["title"] = name

    fig.update_layout(height=300 * len(args), width=600, title_text=title_text)

    fig.show()


def _assign_colors(next_color, color_map, keys):
    colors = []
    for key in keys:
        if key not in color_map:
            color_map[key] = next(next_color)
        colors.append(color_map[key])
    return colors



from datetime import datetime, timedelta

import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1997-01-18T16:40:00,A,,9",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


@pytest.mark.skip(reason="shift to literal not supported")
async def test_shift_to_datetime(source, golden) -> None:
    time = source.col("time")
    shift_to_datetime = datetime(1996, 12, 25, 0, 0, 0)
    golden.jsonl(
        kd.record(
            {"time": time, "shift_to_time_plus_1_day": time.shift_to(shift_to_datetime)}
        )
    )


async def test_shift_to_column(source, golden) -> None:
    time = source.col("time")
    shift_by_timedelta = timedelta(seconds=10)
    golden.jsonl(
        kd.record(
            {
                "time": time,
                "time_plus_seconds": time.shift_to(time.time() + shift_by_timedelta),
            }
        )
    )



from datetime import timedelta

import kaskada as kd
import pyarrow as pa
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1997-01-18T16:40:00,A,,9",
            "1997-01-18T16:40:01,A,3,9",
            "1997-01-18T16:40:03,A,9,1",
            "1997-01-18T16:40:04,A,5,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_time_of_point(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "time_of_m": m.time(),
                "n": n,
                "time_of_n": n.time(),
            }
        )
    )


async def test_time_add_days(source, golden) -> None:
    time = source.col("time")
    golden.jsonl(kd.record({"time": time, "time_plus_day": time + timedelta(days=1)}))


async def test_time_add_hours(source, golden) -> None:
    time = source.col("time")
    golden.jsonl(
        kd.record({"time": time, "time_plus_hours": time + timedelta(hours=1)})
    )


async def test_time_add_minutes(source, golden) -> None:
    time = source.col("time")
    golden.jsonl(
        kd.record({"time": time, "time_plus_minutes": time + timedelta(minutes=1)})
    )


async def test_time_add_days_and_minutes(source, golden) -> None:
    time = source.col("time")
    golden.jsonl(
        kd.record({"time": time, "time_plus_day": time + timedelta(days=3, minutes=1)})
    )


async def test_time_add_seconds(source, golden) -> None:
    time = source.col("time")
    golden.jsonl(
        kd.record({"time": time, "time_plus_seconds": time + timedelta(seconds=5)})
    )


async def test_compare_literal_timedelta(source, golden) -> None:
    time = source.col("time")
    seconds_since = time.seconds_since_previous()
    td = timedelta(milliseconds=1000)
    golden.jsonl(
        kd.record(
            (
                {
                    "seconds_since": seconds_since.cast(pa.int64()),
                    "eq_1s": seconds_since.eq(td),
                    "gt_1s": seconds_since.gt(td),
                    "ge_1s": seconds_since.ge(td),
                    "lt_1s": seconds_since.lt(td),
                    "le_1s": seconds_since.le(td),
                }
            )
        )
    )



import kaskada as kd


async def test_read_jsonl_file(golden) -> None:
    source = await kd.sources.JsonlFile.create(
        "../testdata/purchases/purchases_part1.jsonl",
        time_column="purchase_time",
        key_column="customer_id",
    )
    golden.jsonl(source)

    await source.add_file("../testdata/purchases/purchases_part2.jsonl")
    golden.jsonl(source)


async def test_read_jsonl_file_with_subsort(golden) -> None:
    source = await kd.sources.JsonlFile.create(
        "../testdata/purchases/purchases_part1.jsonl",
        time_column="purchase_time",
        key_column="customer_id",
        subsort_column="subsort_id",
    )
    golden.jsonl(source)

    await source.add_file("../testdata/purchases/purchases_part2.jsonl")
    golden.jsonl(source)



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n,str",
            '1996-12-19T16:39:57,A,5,10,"apple"',
            '1996-12-19T16:39:58,B,24,3,"dog"',
            '1996-12-19T16:39:59,A,17,6,"carrot"',
            "1996-12-19T16:40:00,A,,9,",
            '1996-12-19T16:40:01,A,12,,"eggplant"',
            '1996-12-19T16:40:02,A,,,"fig"',
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_length(source, golden) -> None:
    my_str = source.col("str")
    list = my_str.collect(max=None)
    golden.jsonl(
        kd.record(
            {
                "str": my_str,
                "len_key": my_str.length(),
                "list": list,
                "len_list": list.length(),
            }
        )
    )
